# Copyright (C) 2010-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""A Git repository implementation that uses a Bazaar transport."""

import os
import posixpath
import sys
from io import BytesIO

from dulwich.errors import NoIndexPresent
from dulwich.file import FileLocked, _GitFile
from dulwich.object_store import (PACK_MODE, PACKDIR, PackBasedObjectStore,
                                  read_packs_file)
from dulwich.objects import ShaFile
from dulwich.pack import (PACK_SPOOL_FILE_MAX_SIZE, MemoryPackIndex, Pack,
                          PackData, PackIndexer, PackInflater,
                          PackStreamCopier, compute_file_sha, extend_pack,
                          iter_sha1, load_pack_index_file, write_pack_header,
                          write_pack_index, write_pack_object,
                          write_pack_objects)
from dulwich.refs import SymrefLoop
from dulwich.repo import (BASE_DIRECTORIES, COMMONDIR, CONTROLDIR,
                          INDEX_FILENAME, OBJECTDIR, SYMREF, BaseRepo,
                          InfoRefsContainer, RefsContainer, check_ref_format,
                          read_packed_refs, read_packed_refs_with_peeled,
                          write_packed_refs)

from .. import osutils
from .. import transport as _mod_transport
from .. import ui, urlutils
from ..errors import (AlreadyControlDirError, LockBroken, LockContention,
                      NotLocalUrl, ReadError, TransportNotPossible)
from ..lock import LogicalLockResult
from ..trace import warning
from ..transport import FileExists, NoSuchFile
from ..transport.local import LocalTransport


class _RemoteGitFile(object):

    def __init__(self, transport, filename, mode, bufsize, mask):
        self.transport = transport
        self.filename = filename
        self.mode = mode
        self.bufsize = bufsize
        self.mask = mask
        import tempfile
        self._file = tempfile.SpooledTemporaryFile(max_size=1024 * 1024)
        self._closed = False
        for method in _GitFile.PROXY_METHODS:
            setattr(self, method, getattr(self._file, method))

    def abort(self):
        self._file.close()
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.abort()
        else:
            self.close()

    def __getattr__(self, name):
        """Proxy property calls to the underlying file."""
        if name in _GitFile.PROXY_PROPERTIES:
            return getattr(self._file, name)
        raise AttributeError(name)

    def close(self):
        if self._closed:
            return
        self._file.flush()
        self._file.seek(0)
        self.transport.put_file(self.filename, self._file)
        self._file.close()
        self._closed = True


def TransportGitFile(transport, filename, mode="rb", bufsize=-1, mask=0o644):
    if "a" in mode:
        raise OSError("append mode not supported for Git files")
    if "+" in mode:
        raise OSError("read/write mode not supported for Git files")
    if "b" not in mode:
        raise OSError("text mode not supported for Git files")
    if "w" in mode:
        try:
            return _GitFile(transport.local_abspath(filename), mode, bufsize, mask)
        except NotLocalUrl:
            return _RemoteGitFile(transport, filename, mode, bufsize, mask)
    else:
        return transport.get(filename)


class TransportRefsContainer(RefsContainer):
    """Refs container that reads refs from a transport."""

    def __init__(self, transport, worktree_transport=None):
        self.transport = transport
        if worktree_transport is None:
            worktree_transport = transport
        self.worktree_transport = worktree_transport
        self._packed_refs = None
        self._peeled_refs = None

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.transport)

    def _ensure_dir_exists(self, path):
        self.transport.clone(posixpath.dirname(path)).create_prefix()

    def subkeys(self, base):
        """Refs present in this container under a base.

        :param base: The base to return refs under.
        :return: A set of valid refs in this container under the base; the base
            prefix is stripped from the ref names returned.
        """
        keys = set()
        base_len = len(base) + 1
        for refname in self.allkeys():
            if refname.startswith(base):
                keys.add(refname[base_len:])
        return keys

    def allkeys(self):
        keys = set()
        try:
            self.worktree_transport.get_bytes("HEAD")
        except NoSuchFile:
            pass
        else:
            keys.add(b"HEAD")
        try:
            iter_files = list(self.transport.clone(
                "refs").iter_files_recursive())
            for filename in iter_files:
                unquoted_filename = urlutils.unquote_to_bytes(filename)
                refname = osutils.pathjoin(b"refs", unquoted_filename)
                if check_ref_format(refname):
                    keys.add(refname)
        except (TransportNotPossible, NoSuchFile):
            pass
        keys.update(self.get_packed_refs())
        return keys

    def get_packed_refs(self):
        """Get contents of the packed-refs file.

        :return: Dictionary mapping ref names to SHA1s

        :note: Will return an empty dictionary when no packed-refs file is
            present.
        """
        # TODO: invalidate the cache on repacking
        if self._packed_refs is None:
            # set both to empty because we want _peeled_refs to be
            # None if and only if _packed_refs is also None.
            self._packed_refs = {}
            self._peeled_refs = {}
            try:
                f = self.transport.get("packed-refs")
            except NoSuchFile:
                return {}
            try:
                first_line = next(iter(f)).rstrip()
                if (first_line.startswith(b"# pack-refs") and b" peeled" in
                        first_line):
                    for sha, name, peeled in read_packed_refs_with_peeled(f):
                        self._packed_refs[name] = sha
                        if peeled:
                            self._peeled_refs[name] = peeled
                else:
                    f.seek(0)
                    for sha, name in read_packed_refs(f):
                        self._packed_refs[name] = sha
            finally:
                f.close()
        return self._packed_refs

    def get_peeled(self, name):
        """Return the cached peeled value of a ref, if available.

        :param name: Name of the ref to peel
        :return: The peeled value of the ref. If the ref is known not point to
            a tag, this will be the SHA the ref refers to. If the ref may point
            to a tag, but no cached information is available, None is returned.
        """
        self.get_packed_refs()
        if self._peeled_refs is None or name not in self._packed_refs:
            # No cache: no peeled refs were read, or this ref is loose
            return None
        if name in self._peeled_refs:
            return self._peeled_refs[name]
        else:
            # Known not peelable
            return self[name]

    def read_loose_ref(self, name):
        """Read a reference file and return its contents.

        If the reference file a symbolic reference, only read the first line of
        the file. Otherwise, only read the first 40 bytes.

        :param name: the refname to read, relative to refpath
        :return: The contents of the ref file, or None if the file does not
            exist.
        :raises IOError: if any other error occurs
        """
        if name == b'HEAD':
            transport = self.worktree_transport
        else:
            transport = self.transport
        try:
            f = transport.get(urlutils.quote_from_bytes(name))
        except NoSuchFile:
            return None
        with f:
            try:
                header = f.read(len(SYMREF))
            except ReadError:
                # probably a directory
                return None
            if header == SYMREF:
                # Read only the first line
                return header + next(iter(f)).rstrip(b"\r\n")
            else:
                # Read only the first 40 bytes
                return header + f.read(40 - len(SYMREF))

    def _remove_packed_ref(self, name):
        if self._packed_refs is None:
            return
        # reread cached refs from disk, while holding the lock

        self._packed_refs = None
        self.get_packed_refs()

        if name not in self._packed_refs:
            return

        del self._packed_refs[name]
        if name in self._peeled_refs:
            del self._peeled_refs[name]
        with self.transport.open_write_stream("packed-refs") as f:
            write_packed_refs(f, self._packed_refs, self._peeled_refs)

    def set_symbolic_ref(self, name, other):
        """Make a ref point at another ref.

        :param name: Name of the ref to set
        :param other: Name of the ref to point at
        """
        self._check_refname(name)
        self._check_refname(other)
        if name != b'HEAD':
            transport = self.transport
            self._ensure_dir_exists(urlutils.quote_from_bytes(name))
        else:
            transport = self.worktree_transport
        transport.put_bytes(urlutils.quote_from_bytes(
            name), SYMREF + other + b'\n')

    def set_if_equals(self, name, old_ref, new_ref):
        """Set a refname to new_ref only if it currently equals old_ref.

        This method follows all symbolic references, and can be used to perform
        an atomic compare-and-swap operation.

        :param name: The refname to set.
        :param old_ref: The old sha the refname must refer to, or None to set
            unconditionally.
        :param new_ref: The new sha the refname will refer to.
        :return: True if the set was successful, False otherwise.
        """
        self._check_refname(name)
        try:
            realnames, _ = self.follow(name)
            realname = realnames[-1]
        except (KeyError, IndexError, SymrefLoop):
            realname = name
        if realname == b'HEAD':
            transport = self.worktree_transport
        else:
            transport = self.transport
            self._ensure_dir_exists(urlutils.quote_from_bytes(realname))
        transport.put_bytes(urlutils.quote_from_bytes(
            realname), new_ref + b"\n")
        return True

    def add_if_new(self, name, ref):
        """Add a new reference only if it does not already exist.

        This method follows symrefs, and only ensures that the last ref in the
        chain does not exist.

        :param name: The refname to set.
        :param ref: The new sha the refname will refer to.
        :return: True if the add was successful, False otherwise.
        """
        try:
            realnames, contents = self.follow(name)
            if contents is not None:
                return False
            realname = realnames[-1]
        except (KeyError, IndexError):
            realname = name
        self._check_refname(realname)
        if realname == b'HEAD':
            transport = self.worktree_transport
        else:
            transport = self.transport
            self._ensure_dir_exists(urlutils.quote_from_bytes(realname))
        transport.put_bytes(urlutils.quote_from_bytes(realname), ref + b"\n")
        return True

    def remove_if_equals(self, name, old_ref):
        """Remove a refname only if it currently equals old_ref.

        This method does not follow symbolic references. It can be used to
        perform an atomic compare-and-delete operation.

        :param name: The refname to delete.
        :param old_ref: The old sha the refname must refer to, or None to
            delete unconditionally.
        :return: True if the delete was successful, False otherwise.
        """
        self._check_refname(name)
        # may only be packed
        if name == b'HEAD':
            transport = self.worktree_transport
        else:
            transport = self.transport
        try:
            transport.delete(urlutils.quote_from_bytes(name))
        except NoSuchFile:
            pass
        self._remove_packed_ref(name)
        return True

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def unlock_ref(self, name):
        if name == b"HEAD":
            transport = self.worktree_transport
        else:
            transport = self.transport
        lockname = name + b".lock"
        try:
            transport.delete(urlutils.quote_from_bytes(lockname))
        except NoSuchFile:
            pass

    def lock_ref(self, name):
        if name == b"HEAD":
            transport = self.worktree_transport
        else:
            transport = self.transport
        self._ensure_dir_exists(urlutils.quote_from_bytes(name))
        lockname = urlutils.quote_from_bytes(name + b".lock")
        try:
            local_path = transport.local_abspath(
                urlutils.quote_from_bytes(name))
        except NotLocalUrl:
            # This is racy, but what can we do?
            if transport.has(lockname):
                raise LockContention(name)
            transport.put_bytes(lockname, b'Locked by brz-git')
            return LogicalLockResult(lambda: transport.delete(lockname))
        else:
            try:
                gf = TransportGitFile(transport, urlutils.quote_from_bytes(name), 'wb')
            except FileLocked as e:
                raise LockContention(name, e)
            else:
                def unlock():
                    try:
                        transport.delete(lockname)
                    except NoSuchFile:
                        raise LockBroken(lockname)
                    # GitFile.abort doesn't care if the lock has already
                    # disappeared
                    gf.abort()
                return LogicalLockResult(unlock)


# TODO(jelmer): Use upstream read_gitfile; unfortunately that expects strings
# rather than bytes..
def read_gitfile(f):
    """Read a ``.git`` file.

    The first line of the file should start with "gitdir: "

    :param f: File-like object to read from
    :return: A path
    """
    cs = f.read()
    if not cs.startswith(b"gitdir: "):
        raise ValueError("Expected file to start with 'gitdir: '")
    return cs[len(b"gitdir: "):].rstrip(b"\n")


class TransportRepo(BaseRepo):

    def __init__(self, transport, bare, refs_text=None):
        self.transport = transport
        self.bare = bare
        try:
            with transport.get(CONTROLDIR) as f:
                path = read_gitfile(f)
        except (ReadError, NoSuchFile):
            if self.bare:
                self._controltransport = self.transport
            else:
                self._controltransport = self.transport.clone('.git')
        else:
            self._controltransport = self.transport.clone(
                urlutils.quote_from_bytes(path))
        commondir = self.get_named_file(COMMONDIR)
        if commondir is not None:
            with commondir:
                commondir = os.path.join(
                    self.controldir(),
                    commondir.read().rstrip(b"\r\n").decode(
                        sys.getfilesystemencoding()))
                self._commontransport = \
                    _mod_transport.get_transport_from_path(commondir)
        else:
            self._commontransport = self._controltransport
        config = self.get_config()
        object_store = TransportObjectStore.from_config(
            self._commontransport.clone(OBJECTDIR),
            config)
        if refs_text is not None:
            refs_container = InfoRefsContainer(BytesIO(refs_text))
            try:
                head = TransportRefsContainer(
                    self._commontransport).read_loose_ref(b"HEAD")
            except KeyError:
                pass
            else:
                refs_container._refs[b"HEAD"] = head
        else:
            refs_container = TransportRefsContainer(
                self._commontransport, self._controltransport)
        super().__init__(object_store,
                                            refs_container)

    def controldir(self):
        return self._controltransport.local_abspath('.')

    def commondir(self):
        return self._commontransport.local_abspath('.')

    def close(self):
        """Close any files opened by this repository."""
        self.object_store.close()

    @property
    def path(self):
        return self.transport.local_abspath('.')

    def _determine_file_mode(self):
        # Be consistent with bzr
        if sys.platform == 'win32':
            return False
        return True

    def _determine_symlinks(self):
        try:
            return osutils.supports_symlinks(self.path)
        except NotLocalUrl:
            # TODO(jelmer): Query the transport
            return sys.platform != 'win32'

    def get_named_file(self, path):
        """Get a file from the control dir with a specific name.

        Although the filename should be interpreted as a filename relative to
        the control dir in a disk-baked Repo, the object returned need not be
        pointing to a file in that location.

        :param path: The path to the file, relative to the control dir.
        :return: An open file object, or None if the file does not exist.
        """
        try:
            return self._controltransport.get(path.lstrip('/'))
        except NoSuchFile:
            return None

    def _put_named_file(self, relpath, contents):
        self._controltransport.put_bytes(relpath, contents)

    def index_path(self):
        """Return the path to the index file."""
        return self._controltransport.local_abspath(INDEX_FILENAME)

    def open_index(self):
        """Open the index for this repository."""
        from dulwich.index import Index
        if not self.has_index():
            raise NoIndexPresent()
        return Index(self.index_path())

    def has_index(self):
        """Check if an index is present."""
        # Bare repos must never have index files; non-bare repos may have a
        # missing index file, which is treated as empty.
        return not self.bare

    def get_config(self):
        from dulwich.config import ConfigFile
        try:
            with self._controltransport.get('config') as f:
                return ConfigFile.from_file(f)
        except NoSuchFile:
            return ConfigFile()

    def get_config_stack(self):
        from dulwich.config import StackedConfig
        backends = []
        p = self.get_config()
        if p is not None:
            backends.append(p)
            writable = p
        else:
            writable = None
        backends.extend(StackedConfig.default_backends())
        return StackedConfig(backends, writable=writable)

    def __repr__(self):
        return "<{} for {!r}>".format(self.__class__.__name__, self.transport)

    @classmethod
    def init(cls, transport, bare=False):
        if not bare:
            try:
                transport.mkdir(".git")
            except FileExists:
                raise AlreadyControlDirError(transport.base)
            control_transport = transport.clone(".git")
        else:
            control_transport = transport
        for d in BASE_DIRECTORIES:
            try:
                control_transport.mkdir("/".join(d))
            except FileExists:
                pass
        try:
            control_transport.mkdir(OBJECTDIR)
        except FileExists:
            raise AlreadyControlDirError(transport.base)
        TransportObjectStore.init(control_transport.clone(OBJECTDIR))
        ret = cls(transport, bare)
        ret.refs.set_symbolic_ref(b"HEAD", b"refs/heads/master")
        ret._init_files(bare)
        return ret


class TransportObjectStore(PackBasedObjectStore):
    """Git-style object store that exists on disk."""

    def __init__(self, transport,
                 loose_compression_level=-1, pack_compression_level=-1):
        """Open an object store.

        :param transport: Transport to open data from
        """
        super().__init__()
        self.pack_compression_level = pack_compression_level
        self.loose_compression_level = loose_compression_level
        self.transport = transport
        self.pack_transport = self.transport.clone(PACKDIR)
        self._alternates = None

    @classmethod
    def from_config(cls, path, config):
        try:
            default_compression_level = int(config.get(
                (b'core', ), b'compression').decode())
        except KeyError:
            default_compression_level = -1
        try:
            loose_compression_level = int(config.get(
                (b'core', ), b'looseCompression').decode())
        except KeyError:
            loose_compression_level = default_compression_level
        try:
            pack_compression_level = int(config.get(
                (b'core', ), 'packCompression').decode())
        except KeyError:
            pack_compression_level = default_compression_level
        return cls(path, loose_compression_level, pack_compression_level)

    def __eq__(self, other):
        if not isinstance(other, TransportObjectStore):
            return False
        return self.transport == other.transport

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.transport)

    @property
    def alternates(self):
        if self._alternates is not None:
            return self._alternates
        self._alternates = []
        for path in self._read_alternate_paths():
            # FIXME: Check path
            t = _mod_transport.get_transport_from_path(path)
            self._alternates.append(self.__class__(t))
        return self._alternates

    def _read_alternate_paths(self):
        try:
            f = self.transport.get("info/alternates")
        except NoSuchFile:
            return []
        ret = []
        with f:
            for l in f.read().splitlines():
                if l[0] == b"#":
                    continue
                if os.path.isabs(l):
                    continue
                ret.append(l)
            return ret

    def _update_pack_cache(self):
        pack_files = set(self._pack_names())
        new_packs = []
        for basename in pack_files:
            pack_name = basename + ".pack"
            if basename not in self._pack_cache:
                try:
                    size = self.pack_transport.stat(pack_name).st_size
                except TransportNotPossible:
                    size = None
                pd = PackData(
                    pack_name, self.pack_transport.get(pack_name),
                    size=size)
                idxname = basename + ".idx"
                idx = load_pack_index_file(
                    idxname, self.pack_transport.get(idxname))
                pack = Pack.from_objects(pd, idx)
                pack._basename = basename
                self._pack_cache[basename] = pack
                new_packs.append(pack)
        # Remove disappeared pack files
        for n in set(self._pack_cache) - pack_files:
            self._pack_cache.pop(n).close()
        return new_packs

    def _pack_names(self):
        pack_files = []
        try:
            dir_contents = self.pack_transport.list_dir(".")
            for name in dir_contents:
                if name.startswith("pack-") and name.endswith(".pack"):
                    # verify that idx exists first (otherwise the pack was not yet
                    # fully written)
                    idx_name = os.path.splitext(name)[0] + ".idx"
                    if idx_name in dir_contents:
                        pack_files.append(os.path.splitext(name)[0])
        except TransportNotPossible:
            try:
                f = self.transport.get('info/packs')
            except NoSuchFile:
                warning('No info/packs on remote host;'
                        'run \'git update-server-info\' on remote.')
            else:
                with f:
                    pack_files = [
                        os.path.splitext(name)[0]
                        for name in read_packs_file(f)]
        except NoSuchFile:
            pass
        return pack_files

    def _remove_pack(self, pack):
        self.pack_transport.delete(os.path.basename(pack.index.path))
        self.pack_transport.delete(pack.data.filename)
        try:
            del self._pack_cache[os.path.basename(pack._basename)]
        except KeyError:
            pass

    def _iter_loose_objects(self):
        for base in self.transport.list_dir('.'):
            if len(base) != 2:
                continue
            for rest in self.transport.list_dir(base):
                yield (base + rest).encode(sys.getfilesystemencoding())

    def _split_loose_object(self, sha):
        return (sha[:2], sha[2:])

    def _remove_loose_object(self, sha):
        path = osutils.joinpath(self._split_loose_object(sha))
        self.transport.delete(urlutils.quote_from_bytes(path))

    def _get_loose_object(self, sha):
        path = osutils.joinpath(self._split_loose_object(sha))
        try:
            with self.transport.get(urlutils.quote_from_bytes(path)) as f:
                return ShaFile.from_file(f)
        except NoSuchFile:
            return None

    def add_object(self, obj):
        """Add a single object to this object store.

        :param obj: Object to add
        """
        (dir, file) = self._split_loose_object(obj.id)
        try:
            self.transport.mkdir(urlutils.quote_from_bytes(dir))
        except FileExists:
            pass
        path = urlutils.quote_from_bytes(osutils.pathjoin(dir, file))
        if self.transport.has(path):
            return  # Already there, no need to write again
        # Backwards compatibility with Dulwich < 0.20, which doesn't support
        # the compression_level parameter.
        if self.loose_compression_level not in (-1, None):
            raw_string = obj.as_legacy_object(
                compression_level=self.loose_compression_level)
        else:
            raw_string = obj.as_legacy_object()
        self.transport.put_bytes(path, raw_string)

    @classmethod
    def init(cls, transport):
        try:
            transport.mkdir('info')
        except FileExists:
            pass
        try:
            transport.mkdir(PACKDIR)
        except FileExists:
            pass
        return cls(transport)

    def _complete_pack(self, f, path, num_objects, indexer, progress=None):
        """Move a specific file containing a pack into the pack directory.

        Note: The file should be on the same file system as the
            packs directory.

        Args:
          f: Open file object for the pack.
          path: Path to the pack file.
          indexer: A PackIndexer for indexing the pack.
        """
        entries = []
        for i, entry in enumerate(indexer):
            if progress is not None:
                progress(("generating index: %d/%d\r" % (i, num_objects)).encode('ascii'))
            entries.append(entry)

        pack_sha, extra_entries = extend_pack(
            f, indexer.ext_refs(), get_raw=self.get_raw, compression_level=self.pack_compression_level,
            progress=progress)
        f.flush()
        try:
            fileno = f.fileno()
        except AttributeError:
            pass
        else:
            os.fsync(fileno)

        entries.extend(extra_entries)

        # Move the pack in.
        entries.sort()
        pack_base_name = "pack-" + iter_sha1(entry[0] for entry in entries).decode('ascii')

        for pack in self.packs:
            if osutils.basename(pack._basename) == pack_base_name:
                f.close()
                return pack

        target_pack_name = pack_base_name + ".pack"
        target_pack_index = pack_base_name + ".idx"
        if sys.platform == "win32":
            # Windows might have the target pack file lingering. Attempt
            # removal, silently passing if the target does not exist.
            with suppress(NoSuchFile):
                self.transport.remove(target_pack_name)

        if path:
            f.close()
            self.pack_transport.ensure_base()
            os.rename(path, self.transport.local_abspath(osutils.pathjoin(PACKDIR, target_pack_name)))
        else:
            f.seek(0)
            self.pack_transport.put_file(target_pack_name, f, mode=PACK_MODE)

        # Write the index.
        with TransportGitFile(self.pack_transport, target_pack_index, "wb", mask=PACK_MODE) as index_file:
            write_pack_index(index_file, entries, pack_sha)

        # Add the pack to the store and return it.
        final_pack = Pack.from_objects(
            PackData(target_pack_name, self.pack_transport.get(target_pack_name)),
            load_pack_index_file(target_pack_index, self.pack_transport.get(target_pack_index)))
        final_pack._basename = pack_base_name

        final_pack.check_length_and_checksum()
        self._add_cached_pack(pack_base_name, final_pack)
        return final_pack

    def add_thin_pack(self, read_all, read_some, progress=None):
        """Add a new thin pack to this object store.

        Thin packs are packs that contain deltas with parents that exist
        outside the pack. They should never be placed in the object store
        directly, and always indexed and completed as they are copied.

        Args:
          read_all: Read function that blocks until the number of
            requested bytes are read.
          read_some: Read function that returns at least one byte, but may
            not return the number of bytes requested.
        Returns: A Pack object pointing at the now-completed thin pack in the
            objects/pack directory.
        """
        import tempfile

        try:
            dir = self.transport.local_abspath('.')
        except NotLocalUrl:
            f = tempfile.SpooledTemporaryFile(prefix="tmp_pack_")
            path = None
        else:
            f = tempfile.NamedTemporaryFile(dir=dir, prefix="tmp_pack_", delete=False)
            path = f.name

        try:
            indexer = PackIndexer(f, resolve_ext_ref=self.get_raw)
            copier = PackStreamCopier(read_all, read_some, f, delta_iter=indexer)
            copier.verify(progress=progress)
            if f.name:
                os.chmod(f.name, PACK_MODE)
            return self._complete_pack(f, path, len(copier), indexer, progress=progress)
        except BaseException:
            f.close()
            if path:
                os.remove(path)
            raise

    def add_pack(self):
        """Add a new pack to this object store.

        Returns: Fileobject to write to, a commit function to
            call when the pack is finished and an abort
            function.
        """
        import tempfile

        try:
            dir = self.transport.local_abspath('.')
        except NotLocalUrl:
            f = tempfile.SpooledTemporaryFile(prefix="tmp_pack_")
            path = None
        else:
            f = tempfile.NamedTemporaryFile(dir=dir, prefix="tmp_pack_", delete=False)
            path = f.name

        def commit():
            if f.tell() > 0:
                f.seek(0)
                with PackData(path, f) as pd:
                    indexer = PackIndexer.for_pack_data(pd, resolve_ext_ref=self.get_raw)
                    return self._complete_pack(f, path, len(pd), indexer)
            else:
                abort()
                return None

        def abort():
            f.close()
            if path:
                os.remove(path)

        return f, commit, abort



# Copyright (C) 2009, 2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for controldir implementations - push."""

from breezy.tests.per_controldir import TestCaseWithControlDir

from ...controldir import NoColocatedBranchSupport
from ...errors import LossyPushToSameVCS, NoSuchRevision, TagsNotSupported
from ...revision import NULL_REVISION
from .. import TestNotApplicable


class TestPush(TestCaseWithControlDir):

    def create_simple_tree(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/a'])
        tree.add(['a'])
        rev_1 = tree.commit('one')
        return tree, rev_1

    def test_push_new_branch(self):
        tree, rev_1 = self.create_simple_tree()
        dir = self.make_repository('dir').controldir
        result = dir.push_branch(tree.branch)
        self.assertEqual(tree.branch, result.source_branch)
        self.assertEqual(dir.open_branch().base, result.target_branch.base)
        self.assertEqual(dir.open_branch().base,
                         tree.branch.get_push_location())

    def test_push_to_colocated_active(self):
        tree, rev_1 = self.create_simple_tree()
        dir = self.make_repository('dir').controldir
        try:
            result = dir.push_branch(tree.branch, name='colo')
        except NoColocatedBranchSupport:
            raise TestNotApplicable('no colocated branch support')
        self.assertEqual(tree.branch, result.source_branch)
        self.assertEqual(dir.open_branch(name='colo').base, result.target_branch.base)
        self.assertEqual(dir.open_branch(name='colo').base,
                         tree.branch.get_push_location())

    def test_push_to_colocated_new_inactive(self):
        tree, rev_1 = self.create_simple_tree()
        target_tree = self.make_branch_and_tree('dir')
        rev_o = target_tree.commit('another')
        try:
            result = target_tree.branch.controldir.push_branch(
                tree.branch, name='colo')
        except NoColocatedBranchSupport:
            raise TestNotApplicable('no colocated branch support')
        target_branch = target_tree.branch.controldir.open_branch(
            name='colo')
        self.assertEqual(tree.branch, result.source_branch)
        self.assertEqual(target_tree.last_revision(), rev_o)
        self.assertEqual(target_tree.branch.last_revision(), rev_o)
        self.assertEqual(target_branch.base, result.target_branch.base)
        self.assertEqual(target_branch.base, tree.branch.get_push_location())
        self.assertNotEqual(
            target_branch.controldir.open_branch(name='').name, 'colo')

    def test_push_to_colocated_existing_inactive(self):
        tree, rev_1 = self.create_simple_tree()
        target_tree = self.make_branch_and_tree('dir')
        rev_o = target_tree.commit('another')
        try:
            target_tree.branch.controldir.create_branch(name='colo')
        except NoColocatedBranchSupport:
            raise TestNotApplicable('no colocated branch support')

        try:
            result = target_tree.branch.controldir.push_branch(
                tree.branch, name='colo')
        except NoColocatedBranchSupport:
            raise TestNotApplicable('no colocated branch support')
        target_branch = target_tree.branch.controldir.open_branch(
            name='colo')
        self.assertEqual(tree.branch, result.source_branch)
        self.assertEqual(target_tree.last_revision(), rev_o)
        self.assertEqual(target_tree.branch.last_revision(), rev_o)
        self.assertEqual(target_branch.base, result.target_branch.base)
        self.assertEqual(target_branch.base, tree.branch.get_push_location())

    def test_push_no_such_revision(self):
        tree, rev_1 = self.create_simple_tree()
        dir = self.make_repository('dir').controldir
        self.assertRaises(
            NoSuchRevision, dir.push_branch,
            tree.branch, revision_id=b'idonotexist')

    def test_push_new_branch_fetch_tags(self):
        builder = self.make_branch_builder('from')
        builder.start_series()
        rev_1 = builder.build_snapshot(None, [
            ('add', ('', None, 'directory', '')),
            ('add', ('filename', None, 'file', b'content'))])
        rev_2 = builder.build_snapshot(
            [rev_1], [('modify', ('filename', b'new-content\n'))])
        rev_3 = builder.build_snapshot(
            [rev_1], [('modify', ('filename', b'new-new-content\n'))])
        builder.finish_series()
        branch = builder.get_branch()
        try:
            branch.tags.set_tag('atag', rev_2)
        except TagsNotSupported:
            raise TestNotApplicable('source format does not support tags')

        dir = self.make_repository('target').controldir
        branch.get_config().set_user_option('branch.fetch_tags', True)
        result = dir.push_branch(branch)
        self.assertEqual(
            {rev_1, rev_2, rev_3},
            set(result.source_branch.repository.all_revision_ids()))
        self.assertEqual(
            {'atag': rev_2}, result.source_branch.tags.get_tag_dict())

    def test_push_new_branch_lossy(self):
        tree, rev_1 = self.create_simple_tree()
        dir = self.make_repository('dir').controldir
        self.assertRaises(LossyPushToSameVCS, dir.push_branch,
                          tree.branch, lossy=True)

    def test_push_new_empty(self):
        tree = self.make_branch_and_tree('tree')
        dir = self.make_repository('dir').controldir
        result = dir.push_branch(tree.branch)
        self.assertEqual(tree.branch.base, result.source_branch.base)
        self.assertEqual(dir.open_branch().base,
                         result.target_branch.base)

    def test_push_incremental(self):
        tree, rev1 = self.create_simple_tree()
        dir = self.make_repository('dir').controldir
        dir.push_branch(tree.branch)
        self.build_tree(['tree/b'])
        tree.add(['b'])
        rev_2 = tree.commit('two')
        result = dir.push_branch(tree.branch)
        self.assertEqual(tree.last_revision(),
                         result.branch_push_result.new_revid)
        self.assertEqual(2, result.branch_push_result.new_revno)
        self.assertEqual(tree.branch.base, result.source_branch.base)
        self.assertEqual(dir.open_branch().base, result.target_branch.base)

    def test_push_tag_selector(self):
        tree, rev1 = self.create_simple_tree()
        try:
            tree.branch.tags.set_tag('tag1', rev1)
        except TagsNotSupported:
            raise TestNotApplicable('tags not supported')
        tree.branch.tags.set_tag('tag2', rev1)
        dir = self.make_repository('dir').controldir
        dir.push_branch(tree.branch, tag_selector=lambda x: x == 'tag1')
        self.assertEqual({'tag1': rev1}, dir.open_branch().tags.get_tag_dict())



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import re

from .. import bzr, commands, controldir, errors, option, registry
from ..builtins import cmd_commit
from ..bzr import knitrepo
from ..commands import parse_args
from . import TestCase


def parse(options, args):
    parser = option.get_optparser(options)
    return parser.parse_args(args)


class OptionTests(TestCase):
    """Command-line option tests"""

    def test_parse_args(self):
        """Option parser"""
        # XXX: Using cmd_commit makes these tests overly sensitive to changes
        # to cmd_commit, when they are meant to be about option parsing in
        # general.
        self.assertEqual(
            ([], {'author': [], 'exclude': [], 'fixes': [], 'help': True,
                  'bugs': []}),
            parse_args(cmd_commit(), ['--help']))
        self.assertEqual(
            ([], {'author': [], 'exclude': [], 'fixes': [], 'message': 'biter',
                  'bugs': []}),
            parse_args(cmd_commit(), ['--message=biter']))

    def test_no_more_opts(self):
        """Terminated options"""
        self.assertEqual(
            (['-file-with-dashes'], {
                'author': [], 'exclude': [], 'fixes': [], 'bugs': []}),
            parse_args(cmd_commit(), ['--', '-file-with-dashes']))

    def test_option_help(self):
        """Options have help strings."""
        out, err = self.run_bzr('commit --help')
        self.assertContainsRe(out,
                              r'--file(.|\n)*Take commit message from this file\.')
        self.assertContainsRe(out, r'-h.*--help')

    def test_option_help_global(self):
        """Global options have help strings."""
        out, err = self.run_bzr('help status')
        self.assertContainsRe(out, r'--show-ids.*Show internal object.')

    def test_option_help_global_hidden(self):
        """Hidden global options have no help strings."""
        out, err = self.run_bzr('help log')
        self.assertNotContainsRe(out, r'--message')

    def test_option_arg_help(self):
        """Help message shows option arguments."""
        out, err = self.run_bzr('help commit')
        self.assertEqual(err, '')
        self.assertContainsRe(out, r'--file[ =]MSGFILE')

    def test_unknown_short_opt(self):
        out, err = self.run_bzr('help -r', retcode=3)
        self.assertContainsRe(err, r'no such option')

    def test_set_short_name(self):
        o = option.Option('wiggle')
        o.set_short_name('w')
        self.assertEqual(o.short_name(), 'w')

    def test_allow_dash(self):
        """Test that we can pass a plain '-' as an argument."""
        self.assertEqual((['-']), parse_args(cmd_commit(), ['-'])[0])

    def parse(self, options, args):
        parser = option.get_optparser(options)
        return parser.parse_args(args)

    def test_conversion(self):
        options = [option.Option('hello')]
        opts, args = self.parse(options, ['--no-hello', '--hello'])
        self.assertEqual(True, opts.hello)
        opts, args = self.parse(options, [])
        self.assertFalse(hasattr(opts, 'hello'))
        opts, args = self.parse(options, ['--hello', '--no-hello'])
        self.assertEqual(False, opts.hello)
        options = [option.Option('number', type=int)]
        opts, args = self.parse(options, ['--number', '6'])
        self.assertEqual(6, opts.number)
        self.assertRaises(errors.CommandError, self.parse, options,
                          ['--number'])
        self.assertRaises(errors.CommandError, self.parse, options,
                          ['--no-number'])
        self.assertRaises(errors.CommandError, self.parse, options,
                          ['--number', 'a'])

    def test_is_hidden(self):
        self.assertTrue(option.Option('foo', hidden=True).is_hidden('foo'))
        self.assertFalse(option.Option('foo', hidden=False).is_hidden('foo'))

    def test_registry_conversion(self):
        registry = controldir.ControlDirFormatRegistry()
        bzr.register_metadir(registry, 'one', 'RepositoryFormat7', 'one help')
        bzr.register_metadir(
            registry, 'two', 'RepositoryFormatKnit1', 'two help')
        bzr.register_metadir(registry, 'hidden', 'RepositoryFormatKnit1',
                             'two help', hidden=True)
        registry.set_default('one')
        options = [option.RegistryOption('format', '', registry, str)]
        opts, args = self.parse(options, ['--format', 'one'])
        self.assertEqual({'format': 'one'}, opts)
        opts, args = self.parse(options, ['--format', 'two'])
        self.assertEqual({'format': 'two'}, opts)
        self.assertRaises(option.BadOptionValue, self.parse, options,
                          ['--format', 'three'])
        self.assertRaises(errors.CommandError, self.parse, options,
                          ['--two'])
        options = [option.RegistryOption('format', '', registry, str,
                                         value_switches=True)]
        opts, args = self.parse(options, ['--two'])
        self.assertEqual({'format': 'two'}, opts)
        opts, args = self.parse(options, ['--two', '--one'])
        self.assertEqual({'format': 'one'}, opts)
        opts, args = self.parse(options, ['--two', '--one',
                                          '--format', 'two'])
        self.assertEqual({'format': 'two'}, opts)
        options = [option.RegistryOption('format', '', registry, str,
                                         enum_switch=False)]
        self.assertRaises(errors.CommandError, self.parse, options,
                          ['--format', 'two'])

    def test_override(self):
        options = [option.Option('hello', type=str),
                   option.Option('hi', type=str, param_name='hello')]
        opts, args = self.parse(options, ['--hello', 'a', '--hello', 'b'])
        self.assertEqual('b', opts.hello)
        opts, args = self.parse(options, ['--hello', 'b', '--hello', 'a'])
        self.assertEqual('a', opts.hello)
        opts, args = self.parse(options, ['--hello', 'a', '--hi', 'b'])
        self.assertEqual('b', opts.hello)
        opts, args = self.parse(options, ['--hi', 'b', '--hello', 'a'])
        self.assertEqual('a', opts.hello)

    def test_registry_converter(self):
        options = [option.RegistryOption('format', '',
                                         controldir.format_registry, controldir.format_registry.make_controldir)]
        opts, args = self.parse(options, ['--format', 'knit'])
        self.assertIsInstance(opts.format.repository_format,
                              knitrepo.RepositoryFormatKnit1)

    def test_lazy_registry(self):
        options = [option.RegistryOption('format', '',
                                         lazy_registry=(
                                             'breezy.controldir', 'format_registry'),
                                         converter=str)]
        opts, args = self.parse(options, ['--format', 'knit'])
        self.assertEqual({'format': 'knit'}, opts)
        self.assertRaises(
            option.BadOptionValue, self.parse, options, ['--format', 'BAD'])

    def test_from_kwargs(self):
        my_option = option.RegistryOption.from_kwargs('my-option',
                                                      help='test option', short='be short', be_long='go long')
        self.assertEqual(['my-option'],
                         [x[0] for x in my_option.iter_switches()])
        my_option = option.RegistryOption.from_kwargs('my-option',
                                                      help='test option', title="My option", short='be short',
                                                      be_long='go long', value_switches=True)
        self.assertEqual(['my-option', 'be-long', 'short'],
                         [x[0] for x in my_option.iter_switches()])
        self.assertEqual('test option', my_option.help)

    def test_help(self):
        registry = controldir.ControlDirFormatRegistry()
        bzr.register_metadir(registry, 'one', 'RepositoryFormat7', 'one help')
        bzr.register_metadir(registry, 'two',
                             'breezy.bzr.knitrepo.RepositoryFormatKnit1',
                             'two help',
                             )
        bzr.register_metadir(registry, 'hidden', 'RepositoryFormat7', 'hidden help',
                             hidden=True)
        registry.set_default('one')
        options = [option.RegistryOption('format', 'format help', registry,
                                         str, value_switches=True, title='Formats')]
        parser = option.get_optparser(options)
        value = parser.format_option_help()
        self.assertContainsRe(value, 'format.*format help')
        self.assertContainsRe(value, 'one.*one help')
        self.assertContainsRe(value, 'Formats:\n *--format')
        self.assertNotContainsRe(value, 'hidden help')

    def test_iter_switches(self):
        opt = option.Option('hello', help='fg')
        self.assertEqual(list(opt.iter_switches()),
                         [('hello', None, None, 'fg')])
        opt = option.Option('hello', help='fg', type=int)
        self.assertEqual(list(opt.iter_switches()),
                         [('hello', None, 'ARG', 'fg')])
        opt = option.Option('hello', help='fg', type=int, argname='gar')
        self.assertEqual(list(opt.iter_switches()),
                         [('hello', None, 'GAR', 'fg')])
        registry = controldir.ControlDirFormatRegistry()
        bzr.register_metadir(registry, 'one', 'RepositoryFormat7', 'one help')
        bzr.register_metadir(registry, 'two',
                             'breezy.bzr.knitrepo.RepositoryFormatKnit1',
                             'two help',
                             )
        registry.set_default('one')
        opt = option.RegistryOption('format', 'format help', registry,
                                    value_switches=False)
        self.assertEqual(list(opt.iter_switches()),
                         [('format', None, 'ARG', 'format help')])
        opt = option.RegistryOption('format', 'format help', registry,
                                    value_switches=True)
        self.assertEqual(list(opt.iter_switches()),
                         [('format', None, 'ARG', 'format help'),
                          ('default', None, None, 'one help'),
                          ('one', None, None, 'one help'),
                          ('two', None, None, 'two help'),
                          ])

    def test_option_callback_bool(self):
        "Test booleans get True and False passed correctly to a callback."""
        cb_calls = []

        def cb(option, name, value, parser):
            cb_calls.append((option, name, value, parser))
        options = [option.Option('hello', custom_callback=cb)]
        opts, args = self.parse(options, ['--hello', '--no-hello'])
        self.assertEqual(2, len(cb_calls))
        opt, name, value, parser = cb_calls[0]
        self.assertEqual('hello', name)
        self.assertTrue(value)
        opt, name, value, parser = cb_calls[1]
        self.assertEqual('hello', name)
        self.assertFalse(value)

    def test_option_callback_str(self):
        """Test callbacks work for string options both long and short."""
        cb_calls = []

        def cb(option, name, value, parser):
            cb_calls.append((option, name, value, parser))
        options = [option.Option('hello', type=str, custom_callback=cb,
                                 short_name='h')]
        opts, args = self.parse(options, ['--hello', 'world', '-h', 'mars'])
        self.assertEqual(2, len(cb_calls))
        opt, name, value, parser = cb_calls[0]
        self.assertEqual('hello', name)
        self.assertEqual('world', value)
        opt, name, value, parser = cb_calls[1]
        self.assertEqual('hello', name)
        self.assertEqual('mars', value)


class TestListOptions(TestCase):
    """Tests for ListOption, used to specify lists on the command-line."""

    def parse(self, options, args):
        parser = option.get_optparser(options)
        return parser.parse_args(args)

    def test_list_option(self):
        options = [option.ListOption('hello', type=str)]
        opts, args = self.parse(options, ['--hello=world', '--hello=sailor'])
        self.assertEqual(['world', 'sailor'], opts.hello)

    def test_list_option_with_dash(self):
        options = [option.ListOption('with-dash', type=str)]
        opts, args = self.parse(options, ['--with-dash=world',
                                          '--with-dash=sailor'])
        self.assertEqual(['world', 'sailor'], opts.with_dash)

    def test_list_option_no_arguments(self):
        options = [option.ListOption('hello', type=str)]
        opts, args = self.parse(options, [])
        self.assertEqual([], opts.hello)

    def test_list_option_with_int_type(self):
        options = [option.ListOption('hello', type=int)]
        opts, args = self.parse(options, ['--hello=2', '--hello=3'])
        self.assertEqual([2, 3], opts.hello)

    def test_list_option_with_int_type_can_be_reset(self):
        options = [option.ListOption('hello', type=int)]
        opts, args = self.parse(options, ['--hello=2', '--hello=3',
                                          '--hello=-', '--hello=5'])
        self.assertEqual([5], opts.hello)

    def test_list_option_can_be_reset(self):
        """Passing an option of '-' to a list option should reset the list."""
        options = [option.ListOption('hello', type=str)]
        opts, args = self.parse(
            options, ['--hello=a', '--hello=b', '--hello=-', '--hello=c'])
        self.assertEqual(['c'], opts.hello)

    def test_option_callback_list(self):
        """Test callbacks work for list options."""
        cb_calls = []

        def cb(option, name, value, parser):
            # Note that the value is a reference so copy to keep it
            cb_calls.append((option, name, value[:], parser))
        options = [option.ListOption('hello', type=str, custom_callback=cb)]
        opts, args = self.parse(options, ['--hello=world', '--hello=mars',
                                          '--hello=-'])
        self.assertEqual(3, len(cb_calls))
        opt, name, value, parser = cb_calls[0]
        self.assertEqual('hello', name)
        self.assertEqual(['world'], value)
        opt, name, value, parser = cb_calls[1]
        self.assertEqual('hello', name)
        self.assertEqual(['world', 'mars'], value)
        opt, name, value, parser = cb_calls[2]
        self.assertEqual('hello', name)
        self.assertEqual([], value)

    def test_list_option_param_name(self):
        """Test list options can have their param_name set."""
        options = [option.ListOption('hello', type=str, param_name='greeting')]
        opts, args = self.parse(
            options, ['--hello=world', '--hello=sailor'])
        self.assertEqual(['world', 'sailor'], opts.greeting)


class TestOptionDefinitions(TestCase):
    """Tests for options in the Bazaar codebase."""

    def get_builtin_command_options(self):
        g = []
        commands.install_bzr_command_hooks()
        for cmd_name in sorted(commands.builtin_command_names()):
            cmd = commands.get_cmd_object(cmd_name)
            for opt_name, opt in sorted(cmd.options().items()):
                g.append((cmd_name, opt))
        self.assertTrue(g)
        return g

    def test_option_grammar(self):
        msgs = []
        # Option help should be written in sentence form, and have a final
        # period with an optional bracketed suffix. All the text should be on
        # one line, because the display code will wrap it.
        option_re = re.compile(r'^[A-Z][^\n]+\.(?: \([^\n]+\))?$')
        for scope, opt in self.get_builtin_command_options():
            for name, _, _, helptxt in opt.iter_switches():
                if name != opt.name:
                    name = "/".join([opt.name, name])
                if not helptxt:
                    msgs.append('%-16s %-16s %s' %
                                ((scope or 'GLOBAL'), name, 'NO HELP'))
                elif not option_re.match(helptxt):
                    msgs.append('%-16s %-16s %s' %
                                ((scope or 'GLOBAL'), name, helptxt))
        if msgs:
            self.fail("The following options don't match the style guide:\n"
                      + '\n'.join(msgs))


class TestOptionMisc(TestCase):

    def test_is_hidden(self):
        registry = controldir.ControlDirFormatRegistry()
        bzr.register_metadir(registry, 'hidden', 'HiddenFormat',
                             'hidden help text', hidden=True)
        bzr.register_metadir(registry, 'visible', 'VisibleFormat',
                             'visible help text', hidden=False)
        format = option.RegistryOption('format', '', registry, str)
        self.assertTrue(format.is_hidden('hidden'))
        self.assertFalse(format.is_hidden('visible'))

    def test_short_name(self):
        registry = controldir.ControlDirFormatRegistry()
        opt = option.RegistryOption('format', help='', registry=registry)
        self.assertEqual(None, opt.short_name())
        opt = option.RegistryOption('format', short_name='F', help='',
                                    registry=registry)
        self.assertEqual('F', opt.short_name())

    def test_option_custom_help(self):
        the_opt = option.Option.OPTIONS['help']
        orig_help = the_opt.help[:]
        my_opt = option.custom_help('help', 'suggest lottery numbers')
        # Confirm that my_opt has my help and the original is unchanged
        self.assertEqual('suggest lottery numbers', my_opt.help)
        self.assertEqual(orig_help, the_opt.help)

    def test_short_value_switches(self):
        reg = registry.Registry()
        reg.register('short', 'ShortChoice')
        reg.register('long', 'LongChoice')
        ropt = option.RegistryOption('choice', '', reg, value_switches=True,
                                     short_value_switches={'short': 's'})
        opts, args = parse([ropt], ['--short'])
        self.assertEqual('ShortChoice', opts.choice)
        opts, args = parse([ropt], ['-s'])
        self.assertEqual('ShortChoice', opts.choice)


class TestVerboseQuietLinkage(TestCase):

    def check(self, parser, level, args):
        option._verbosity_level = 0
        opts, args = parser.parse_args(args)
        self.assertEqual(level, option._verbosity_level)

    def test_verbose_quiet_linkage(self):
        parser = option.get_optparser(
            [v for k, v in sorted(option.Option.STD_OPTIONS.items())])
        self.check(parser, 0, [])
        self.check(parser, 1, ['-v'])
        self.check(parser, 2, ['-v', '-v'])
        self.check(parser, -1, ['-q'])
        self.check(parser, -2, ['-qq'])
        self.check(parser, -1, ['-v', '-v', '-q'])
        self.check(parser, 2, ['-q', '-v', '-v'])
        self.check(parser, 0, ['--no-verbose'])
        self.check(parser, 0, ['-v', '-q', '--no-quiet'])



# Copyright (C) 2005-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__docformat__ = "google"

"""Abstraction for interacting with the user.

Applications can choose different types of UI, and they deal with displaying
messages or progress to the user, and with gathering different types of input.

Several levels are supported, and you can also register new factories such as
for a GUI.

breezy.ui.UIFactory

    Semi-abstract base class

breezy.ui.SilentUIFactory

    Produces no output and cannot take any input; useful for programs using
    breezy in batch mode or for programs such as loggerhead.

breezy.ui.CannedInputUIFactory

    For use in testing; the input values to be returned are provided
    at construction.

breezy.ui.text.TextUIFactory

    Standard text command-line interface, with stdin, stdout, stderr.
    May make more or less advanced use of them, eg in drawing progress bars,
    depending on the detected capabilities of the terminal.
    GUIs may choose to subclass this so that unimplemented methods fall
    back to working through the terminal.
"""

import warnings

_valid_boolean_strings = dict(yes=True, no=False,
                              y=True, n=False,
                              on=True, off=False,
                              true=True, false=False)
_valid_boolean_strings['1'] = True
_valid_boolean_strings['0'] = False


def bool_from_string(s, accepted_values=None):
    """Returns a boolean if the string can be interpreted as such.

    Interpret case insensitive strings as booleans. The default values
    includes: 'yes', 'no, 'y', 'n', 'true', 'false', '0', '1', 'on',
    'off'. Alternative values can be provided with the 'accepted_values'
    parameter.

    Args:
      s: A string that should be interpreted as a boolean. It should be of
         type string or unicode.

      accepted_values: An optional dict with accepted strings as keys and
         True/False as values. The strings will be tested against a lowered
         version of 's'.

    Returns:
      True or False for accepted strings, None otherwise.
    """
    if accepted_values is None:
        accepted_values = _valid_boolean_strings
    val = None
    if isinstance(s, str):
        try:
            val = accepted_values[s.lower()]
        except KeyError:
            pass
    return val


class ConfirmationUserInterfacePolicy:
    """Wrapper for a UIFactory that allows or denies all confirmed actions."""

    def __init__(self, wrapped_ui, default_answer, specific_answers):
        """Generate a proxy UI that does no confirmations.

        Args:
          wrapped_ui: Underlying UIFactory.
          default_answer: Bool for whether requests for
            confirmation from the user should be noninteractively accepted or
            denied.
          specific_answers: Map from confirmation_id to bool answer.
        """
        self.wrapped_ui = wrapped_ui
        self.default_answer = default_answer
        self.specific_answers = specific_answers

    def __getattr__(self, name):
        return getattr(self.wrapped_ui, name)

    def __repr__(self):
        return '{}({!r}, {!r}, {!r})'.format(
            self.__class__.__name__,
            self.wrapped_ui,
            self.default_answer,
            self.specific_answers)

    def confirm_action(self, prompt, confirmation_id, prompt_kwargs):
        if confirmation_id in self.specific_answers:
            return self.specific_answers[confirmation_id]
        elif self.default_answer is not None:
            return self.default_answer
        else:
            return self.wrapped_ui.confirm_action(
                prompt, confirmation_id, prompt_kwargs)


class UIFactory:
    """UI abstraction.

    This tells the library how to display things to the user.  Through this
    layer different applications can choose the style of UI.

    UI Factories are also context managers, for some syntactic sugar some users
    need.

    Attributes:
      suppressed_warnings: Identifiers for user warnings that should
                           no be emitted.
    """

    _user_warning_templates = dict(
        cross_format_fetch=("Doing on-the-fly conversion from "
                            "%(from_format)s to %(to_format)s.\n"
                            "This may take some time. Upgrade the repositories to the "
                            "same format for better performance."
                            ),
        experimental_format_fetch=("Fetching into experimental format "
                                   "%(to_format)s.\n"
                                   "This format may be unreliable or change in the future "
                                   "without an upgrade path.\n"),
        deprecated_command=(
            "The command 'brz %(deprecated_name)s' "
            "has been deprecated in brz %(deprecated_in_version)s. "
            "Please use 'brz %(recommended_name)s' instead."),
        deprecated_command_option=(
            "The option '%(deprecated_name)s' to 'brz %(command)s' "
            "has been deprecated in brz %(deprecated_in_version)s. "
            "Please use '%(recommended_name)s' instead."),
        recommend_upgrade=("%(current_format_name)s is deprecated "
                           "and a better format is available.\n"
                           "It is recommended that you upgrade by "
                           "running the command\n"
                           "  brz upgrade %(basedir)s"),
        locks_steal_dead=(
            "Stole dead lock %(lock_url)s %(other_holder_info)s."),
        not_checking_ssl_cert=(
            "Not checking SSL certificate for %(host)s."),
        )

    def __init__(self):
        self._task_stack = []
        self.suppressed_warnings = set()
        self._quiet = False

    def __enter__(self):
        """Context manager entry support.

        Override in a concrete factory class if initialisation before use is
        needed.
        """
        return self  # This is bound to the 'as' clause in a with statement.

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit support.

        Override in a concrete factory class if more cleanup than a simple
        self.clear_term() is needed when the UIFactory is finished with.
        """
        self.clear_term()
        return False  # propogate exceptions.

    def be_quiet(self, state):
        """Tell the UI to be more quiet, or not.

        Typically this suppresses progress bars; the application may also look
        at ui_factory.is_quiet().
        """
        self._quiet = state

    def confirm_action(self, prompt, confirmation_id, prompt_kwargs):
        """Seek user confirmation for an action.

        If the UI is noninteractive, or the user does not want to be asked
        about this action, True is returned, indicating brz should just
        proceed.

        The confirmation id allows the user to configure certain actions to
        always be confirmed or always denied, and for UIs to specialize the
        display of particular confirmations.

        Args:
          prompt: Suggested text to display to the user.
          prompt_kwargs: A dictionary of arguments that can be
              string-interpolated into the prompt.
          confirmation_id: Unique string identifier for the confirmation.
        """
        return self.get_boolean(prompt % prompt_kwargs)

    def get_password(self, prompt='', **kwargs):
        """Prompt the user for a password.

        Args:
          prompt: The prompt to present the user (must be unicode)
          kwargs: Arguments which will be expanded into the prompt.
                       This lets front ends display different things if
                       they so choose.

        Returns: The password string, return None if the user canceled the
                 request. Note that we do not touch the encoding, users may
                 have whatever they see fit and the password should be
                 transported as is.
        """
        raise NotImplementedError(self.get_password)

    def is_quiet(self):
        return self._quiet

    def make_output_stream(self, encoding=None, encoding_type='replace'):
        """Get a stream for sending out bulk text data.

        This is used for commands that produce bulk text, such as log or diff
        output, as opposed to user interaction.  This should work even for
        non-interactive user interfaces.  Typically this goes to a decorated
        version of stdout, but in a GUI it might be appropriate to send it to a
        window displaying the text.

        The caller may flush but should not close the returned stream.

        Args:
          encoding: Unicode encoding for output; if not specified
            uses the configured 'output_encoding' if any; otherwise the
            terminal encoding.
            (See get_terminal_encoding.)

          encoding_type: How to handle encoding errors:
            replace/strict/escape/exact.  Default is replace.
        """
        out_stream = self._make_output_stream_explicit(encoding, encoding_type)
        return out_stream

    def _make_output_stream_explicit(self, encoding, encoding_type):
        raise NotImplementedError("%s doesn't support make_output_stream"
                                  % (self.__class__.__name__))

    def nested_progress_bar(self):
        """Return a nested progress bar.

        When the bar has been finished with, it should be released by calling
        bar.finished().
        """
        from ..progress import ProgressTask
        if self._task_stack:
            t = ProgressTask(self._task_stack[-1], self)
        else:
            t = ProgressTask(None, self)
        self._task_stack.append(t)
        return t

    def _progress_finished(self, task):
        """Called by the ProgressTask when it finishes"""
        if not self._task_stack:
            warnings.warn("%r finished but nothing is active"
                          % (task,))
        if task in self._task_stack:
            self._task_stack.remove(task)
        else:
            warnings.warn("%r is not in active stack %r"
                          % (task, self._task_stack))
        if not self._task_stack:
            self._progress_all_finished()

    def _progress_all_finished(self):
        """Called when the top-level progress task finished"""
        pass

    def _progress_updated(self, task):
        """Called by the ProgressTask when it changes.

        Should be specialized to draw the progress.
        """
        pass

    def clear_term(self):
        """Prepare the terminal for output.

        This will, for example, clear text progress bars, and leave the
        cursor at the leftmost position.
        """
        pass

    def format_user_warning(self, warning_id, message_args):
        try:
            template = self._user_warning_templates[warning_id]
        except KeyError:
            fail = "brz warning: {!r}, {!r}".format(warning_id, message_args)
            warnings.warn("no template for warning: "
                          + fail)   # so tests will fail etc
            return str(fail)
        try:
            return str(template) % message_args
        except ValueError as e:
            fail = "brz unprintable warning: {!r}, {!r}, {}".format(
                warning_id, message_args, e)
            warnings.warn(fail)   # so tests will fail etc
            return str(fail)

    def choose(self, msg, choices, default=None):
        """Prompt the user for a list of alternatives.

        Args:
          msg: message to be shown as part of the prompt.

          choices: list of choices, with the individual choices separated
            by '\n', e.g.: choose("Save changes?", "&Yes\n&No\n&Cancel"). The
            letter after the '&' is the shortcut key for that choice. Thus you
            can type 'c' to select "Cancel".  Shorcuts are case insensitive.
            The shortcut does not need to be the first letter. If a shorcut key
            is not provided, the first letter for the choice will be used.

          default: default choice (index), returned for example when enter
            is pressed for the console version.

        Returns:
          the index fo the user choice (so '0', '1' or '2' for respectively
          yes/no/cancel in the previous example).
        """
        raise NotImplementedError(self.choose)

    def get_boolean(self, prompt):
        """Get a boolean question answered from the user.

        Args:
          prompt: a message to prompt the user with. Should be a single
            line without terminating \\n.

        Returns:
          True or False for y/yes or n/no.
        """
        choice = self.choose(prompt + '?', '&yes\n&no', default=None)
        return 0 == choice

    def get_integer(self, prompt):
        """Get an integer from the user.

        Args:
          prompt: a message to prompt the user with. Could be a multi-line
            prompt but without a terminating \\n.

        Returns:
          A signed integer.
        """
        raise NotImplementedError(self.get_integer)

    def make_progress_view(self):
        """Construct a new ProgressView object for this UI.

        Application code should normally not call this but instead
        nested_progress_bar().
        """
        return NullProgressView()

    def recommend_upgrade(self, current_format_name, basedir):
        """Recommend the user upgrade a control directory.

        Args:
          current_format_name: Description of the current format
          basedir: Location of the control dir
        """
        self.show_user_warning('recommend_upgrade',
                               current_format_name=current_format_name, basedir=basedir)

    def report_transport_activity(self, transport, byte_count, direction):
        """Called by transports as they do IO.

        This may update a progress bar, spinner, or similar display.
        By default it does nothing.
        """
        pass

    def log_transport_activity(self, display=False):
        """Write out whatever transport activity has been measured.

        Implementations are allowed to do nothing, but it is useful if they can
        write a line to the log file.

        Args:
          display: If False, only log to disk, if True also try to display
            a message to the user.

        Returns:
          None
        """
        # Default implementation just does nothing
        pass

    def show_user_warning(self, warning_id, **message_args):
        """Show a warning to the user.

        This is specifically for things that are under the user's control (eg
        outdated formats), not for internal program warnings like deprecated
        APIs.

        This can be overridden by UIFactory subclasses to show it in some
        appropriate way; the default UIFactory is noninteractive and does
        nothing.  format_user_warning maps it to a string, though other
        presentations can be used for particular UIs.

        Args:
          warning_id: An identifier like 'cross_format_fetch' used to
            check if the message is suppressed and to look up the string.
          message_args: Arguments to be interpolated into the message.
        """
        pass

    def show_error(self, msg):
        """Show an error message (not an exception) to the user.

        The message should not have an error prefix or trailing newline.  That
        will be added by the factory if appropriate.
        """
        raise NotImplementedError(self.show_error)

    def show_message(self, msg):
        """Show a message to the user."""
        raise NotImplementedError(self.show_message)

    def show_warning(self, msg):
        """Show a warning to the user."""
        raise NotImplementedError(self.show_warning)


class NoninteractiveUIFactory(UIFactory):
    """Base class for UIs with no user."""

    def confirm_action(self, prompt, confirmation_id, prompt_kwargs):
        return True

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)


class SilentUIFactory(NoninteractiveUIFactory):
    """A UI Factory which never prints anything.

    This is the default UI, if another one is never registered by a program
    using breezy, and it's also active for example inside 'brz serve'.

    Methods that try to read from the user raise an error; methods that do
    output do nothing.
    """

    def __init__(self):
        UIFactory.__init__(self)

    def note(self, msg):
        pass

    def get_username(self, prompt, **kwargs):
        return None

    def _make_output_stream_explicit(self, encoding, encoding_type):
        return NullOutputStream(encoding)

    def show_error(self, msg):
        pass

    def show_message(self, msg):
        pass

    def show_warning(self, msg):
        pass


class CannedInputUIFactory(SilentUIFactory):
    """A silent UI that return canned input."""

    def __init__(self, responses):
        self.responses = responses

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.responses)

    def confirm_action(self, prompt, confirmation_id, args):
        return self.get_boolean(prompt % args)

    def get_boolean(self, prompt):
        return self.responses.pop(0)

    def get_integer(self, prompt):
        return self.responses.pop(0)

    def get_password(self, prompt='', **kwargs):
        return self.responses.pop(0)

    def get_username(self, prompt, **kwargs):
        return self.responses.pop(0)

    def assert_all_input_consumed(self):
        if self.responses:
            raise AssertionError("expected all input in %r to be consumed"
                                 % (self,))


ui_factory = SilentUIFactory()
# IMPORTANT: never import this symbol directly. ONLY ever access it as
# ui.ui_factory, so that you refer to the current value.


def make_ui_for_terminal(stdin, stdout, stderr):
    """Construct and return a suitable UIFactory for a text mode program.
    """
    # this is now always TextUIFactory, which in turn decides whether it
    # should display progress bars etc
    from .text import TextUIFactory
    return TextUIFactory(stdin, stdout, stderr)


class NullProgressView:
    """Soak up and ignore progress information."""

    def clear(self):
        pass

    def show_progress(self, task):
        pass

    def show_transport_activity(self, transport, direction, byte_count):
        pass

    def log_transport_activity(self, display=False):
        pass


class NullOutputStream:
    """Acts like a file, but discard all output."""

    def __init__(self, encoding):
        self.encoding = encoding

    def write(self, data):
        pass

    def writelines(self, data):
        pass

    def close(self):
        pass



# Copyright (C) 2005, 2006, 2007, 2009, 2011, 2012, 2013, 2016 Canonical Ltd
#   Authors: Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""GPG signing and checking logic."""

import os
from typing import Dict, List, Optional

from . import config, errors, trace, ui
from .i18n import gettext, ngettext

# verification results
SIGNATURE_VALID = 0
SIGNATURE_KEY_MISSING = 1
SIGNATURE_NOT_VALID = 2
SIGNATURE_NOT_SIGNED = 3
SIGNATURE_EXPIRED = 4

MODE_NORMAL = 0
MODE_DETACH = 1
MODE_CLEAR = 2


class GpgNotInstalled(errors.DependencyNotPresent):

    _fmt = ('python-gpg is not installed, it is needed to create or '
            'verify signatures. %(error)s')

    def __init__(self, error):
        errors.DependencyNotPresent.__init__(self, 'gpg', error)


class SigningFailed(errors.BzrError):

    _fmt = 'Failed to GPG sign data: "%(error)s"'

    def __init__(self, error):
        errors.BzrError.__init__(self, error=error)


class SignatureVerificationFailed(errors.BzrError):

    _fmt = 'Failed to verify GPG signature data with error "%(error)s"'

    def __init__(self, error):
        errors.BzrError.__init__(self, error=error)


def bulk_verify_signatures(repository, revids, strategy,
                           process_events_callback=None):
    """Do verifications on a set of revisions

    :param repository: repository object
    :param revids: list of revision ids to verify
    :param strategy: GPG strategy to use
    :param process_events_callback: method to call for GUI frontends that
        want to keep their UI refreshed

    :return: count dictionary of results of each type,
             result list for each revision,
             boolean True if all results are verified successfully
    """
    count = {SIGNATURE_VALID: 0,
             SIGNATURE_KEY_MISSING: 0,
             SIGNATURE_NOT_VALID: 0,
             SIGNATURE_NOT_SIGNED: 0,
             SIGNATURE_EXPIRED: 0}
    result = []
    all_verifiable = True
    total = len(revids)
    with ui.ui_factory.nested_progress_bar() as pb:
        for i, (rev_id, verification_result, uid) in enumerate(
                repository.verify_revision_signatures(
                    revids, strategy)):
            pb.update("verifying signatures", i, total)
            result.append([rev_id, verification_result, uid])
            count[verification_result] += 1
            if verification_result != SIGNATURE_VALID:
                all_verifiable = False
            if process_events_callback is not None:
                process_events_callback()
    return (count, result, all_verifiable)


class DisabledGPGStrategy:
    """A GPG Strategy that makes everything fail."""

    @staticmethod
    def verify_signatures_available():
        return True

    def __init__(self, ignored):
        """Real strategies take a configuration."""

    def sign(self, content, mode):
        raise SigningFailed('Signing is disabled.')

    def verify(self, signed_data, signature=None):
        raise SignatureVerificationFailed('Signature verification is \
disabled.')

    def set_acceptable_keys(self, command_line_input):
        pass


class LoopbackGPGStrategy:
    """A GPG Strategy that acts like 'cat' - data is just passed through.
    Used in tests.
    """

    @staticmethod
    def verify_signatures_available():
        return True

    def __init__(self, ignored):
        """Real strategies take a configuration."""

    def sign(self, content, mode):
        return (b"-----BEGIN PSEUDO-SIGNED CONTENT-----\n" + content
                + b"-----END PSEUDO-SIGNED CONTENT-----\n")

    def verify(self, signed_data, signature=None):
        plain_text = signed_data.replace(
            b"-----BEGIN PSEUDO-SIGNED CONTENT-----\n", b"")
        plain_text = plain_text.replace(
            b"-----END PSEUDO-SIGNED CONTENT-----\n", b"")
        return SIGNATURE_VALID, None, plain_text

    def set_acceptable_keys(self, command_line_input):
        if command_line_input is not None:
            patterns = command_line_input.split(",")
            self.acceptable_keys = []
            for pattern in patterns:
                if pattern == "unknown":
                    pass
                else:
                    self.acceptable_keys.append(pattern)


def _set_gpg_tty():
    tty = os.environ.get('TTY')
    if tty is not None:
        os.environ['GPG_TTY'] = tty
        trace.mutter('setting GPG_TTY=%s', tty)
    else:
        # This is not quite worthy of a warning, because some people
        # don't need GPG_TTY to be set. But it is worthy of a big mark
        # in brz.log, so that people can debug it if it happens to them
        trace.mutter('** Env var TTY empty, cannot set GPG_TTY.'
                     '  Is TTY exported?')


class GPGStrategy:
    """GPG Signing and checking facilities."""

    acceptable_keys: Optional[List[str]] = None

    def __init__(self, config_stack):
        self._config_stack = config_stack
        try:
            import gpg
            self.context = gpg.Context()
            self.context.armor = True
            self.context.signers = self._get_signing_keys()
        except ModuleNotFoundError:
            pass  # can't use verify()

    def _get_signing_keys(self):
        import gpg
        keyname = self._config_stack.get('gpg_signing_key')
        if keyname == 'default':
            # Leave things to gpg
            return []

        if keyname:
            try:
                return [self.context.get_key(keyname)]
            except gpg.errors.KeyNotFound:
                pass

        if keyname is None:
            # not setting gpg_signing_key at all means we should
            # use the user email address
            keyname = config.extract_email_address(
                self._config_stack.get('email'))
        if keyname == 'default':
            return []
        possible_keys = self.context.keylist(keyname, secret=True)
        try:
            return [next(possible_keys)]
        except StopIteration:
            return []

    @staticmethod
    def verify_signatures_available():
        """
        check if this strategy can verify signatures

        :return: boolean if this strategy can verify signatures
        """
        try:
            import gpg  # noqa: F401
            return True
        except ModuleNotFoundError:
            return False

    def sign(self, content, mode):
        try:
            import gpg
        except ModuleNotFoundError as error:
            raise GpgNotInstalled(
                'Set create_signatures=no to disable creating signatures.')

        if isinstance(content, str):
            raise errors.BzrBadParameterUnicode('content')

        plain_text = gpg.Data(content)
        try:
            output, result = self.context.sign(
                plain_text, mode={
                    MODE_DETACH: gpg.constants.sig.mode.DETACH,
                    MODE_CLEAR: gpg.constants.sig.mode.CLEAR,
                    MODE_NORMAL: gpg.constants.sig.mode.NORMAL,
                    }[mode])
        except gpg.errors.GPGMEError as error:
            raise SigningFailed(str(error))
        except gpg.errors.InvalidSigners as error:
            raise SigningFailed(str(error))

        return output

    def verify(self, signed_data, signature=None):
        """Check content has a valid signature.

        :param signed_data; Signed data
        :param signature: optional signature (if detached)

        :return: SIGNATURE_VALID or a failed SIGNATURE_ value, key uid if valid, plain text
        """
        try:
            import gpg
        except ModuleNotFoundError as error:
            raise GpgNotInstalled(
                'Set check_signatures=ignore to disable verifying signatures.')

        signed_data = gpg.Data(signed_data)
        if signature:
            signature = gpg.Data(signature)
        try:
            plain_output, result = self.context.verify(signed_data, signature)
        except gpg.errors.BadSignatures as error:
            fingerprint = error.result.signatures[0].fpr
            if error.result.signatures[0].summary & gpg.constants.SIGSUM_KEY_EXPIRED:
                expires = self.context.get_key(
                    error.result.signatures[0].fpr).subkeys[0].expires
                if expires > error.result.signatures[0].timestamp:
                    # The expired key was not expired at time of signing.
                    # test_verify_expired_but_valid()
                    return SIGNATURE_EXPIRED, fingerprint[-8:], None
                else:
                    # I can't work out how to create a test where the signature
                    # was expired at the time of signing.
                    return SIGNATURE_NOT_VALID, None, None

            # GPG does not know this key.
            # test_verify_unknown_key()
            if (error.result.signatures[0].summary &
                    gpg.constants.SIGSUM_KEY_MISSING):
                return SIGNATURE_KEY_MISSING, fingerprint[-8:], None

            return SIGNATURE_NOT_VALID, None, None
        except gpg.errors.GPGMEError as error:
            raise SignatureVerificationFailed(error)

        # No result if input is invalid.
        # test_verify_invalid()
        if len(result.signatures) == 0:
            return SIGNATURE_NOT_VALID, None, plain_output

        # User has specified a list of acceptable keys, check our result is in
        # it.  test_verify_unacceptable_key()
        fingerprint = result.signatures[0].fpr
        if self.acceptable_keys is not None:
            if fingerprint not in self.acceptable_keys:
                return SIGNATURE_KEY_MISSING, fingerprint[-8:], plain_output
        # Yay gpg set the valid bit.
        # Can't write a test for this one as you can't set a key to be
        # trusted using gpg.
        if result.signatures[0].summary & gpg.constants.SIGSUM_VALID:
            key = self.context.get_key(fingerprint)
            name = key.uids[0].name
            if isinstance(name, bytes):
                name = name.decode('utf-8')
            email = key.uids[0].email
            if isinstance(email, bytes):
                email = email.decode('utf-8')
            return (SIGNATURE_VALID, name + " <" + email + ">", plain_output)
        # Sigsum_red indicates a problem, unfortunatly I have not been able
        # to write any tests which actually set this.
        if result.signatures[0].summary & gpg.constants.SIGSUM_RED:
            return SIGNATURE_NOT_VALID, None, plain_output
        # Summary isn't set if sig is valid but key is untrusted but if user
        # has explicity set the key as acceptable we can validate it.
        if (result.signatures[0].summary == 0 and
                self.acceptable_keys is not None):
            if fingerprint in self.acceptable_keys:
                # test_verify_untrusted_but_accepted()
                return SIGNATURE_VALID, None, plain_output
        # test_verify_valid_but_untrusted()
        if result.signatures[0].summary == 0 and self.acceptable_keys is None:
            return SIGNATURE_NOT_VALID, None, plain_output
        # Other error types such as revoked keys should (I think) be caught by
        # SIGSUM_RED so anything else means something is buggy.
        raise SignatureVerificationFailed(
            "Unknown GnuPG key verification result")

    def set_acceptable_keys(self, command_line_input):
        """Set the acceptable keys for verifying with this GPGStrategy.

        :param command_line_input: comma separated list of patterns from
                                command line
        :return: nothing
        """
        patterns = None
        acceptable_keys_config = self._config_stack.get('acceptable_keys')
        if acceptable_keys_config is not None:
            patterns = acceptable_keys_config
        if command_line_input is not None:  # command line overrides config
            patterns = command_line_input.split(',')

        if patterns:
            self.acceptable_keys = []
            for pattern in patterns:
                result = self.context.keylist(pattern)
                found_key = False
                for key in result:
                    found_key = True
                    self.acceptable_keys.append(key.subkeys[0].fpr)
                    trace.mutter("Added acceptable key: " + key.subkeys[0].fpr)
                if not found_key:
                    trace.note(gettext(
                        "No GnuPG key results for pattern: {0}"
                        ).format(pattern))


def valid_commits_message(count):
    """returns message for number of commits"""
    return gettext("{0} commits with valid signatures").format(
        count[SIGNATURE_VALID])


def unknown_key_message(count):
    """returns message for number of commits"""
    return ngettext("{0} commit with unknown key",
                    "{0} commits with unknown keys",
                    count[SIGNATURE_KEY_MISSING]).format(
        count[SIGNATURE_KEY_MISSING])


def commit_not_valid_message(count):
    """returns message for number of commits"""
    return ngettext("{0} commit not valid",
                    "{0} commits not valid",
                    count[SIGNATURE_NOT_VALID]).format(
        count[SIGNATURE_NOT_VALID])


def commit_not_signed_message(count):
    """returns message for number of commits"""
    return ngettext("{0} commit not signed",
                    "{0} commits not signed",
                    count[SIGNATURE_NOT_SIGNED]).format(
        count[SIGNATURE_NOT_SIGNED])


def expired_commit_message(count):
    """returns message for number of commits"""
    return ngettext("{0} commit with key now expired",
                    "{0} commits with key now expired",
                    count[SIGNATURE_EXPIRED]).format(
        count[SIGNATURE_EXPIRED])


def verbose_expired_key_message(result, repo) -> List[str]:
    """takes a verify result and returns list of expired key info"""
    signers: Dict[str, int] = {}
    fingerprint_to_authors = {}
    for rev_id, validity, fingerprint in result:
        if validity == SIGNATURE_EXPIRED:
            revision = repo.get_revision(rev_id)
            authors = ', '.join(revision.get_apparent_authors())
            signers.setdefault(fingerprint, 0)
            signers[fingerprint] += 1
            fingerprint_to_authors[fingerprint] = authors
    ret: List[str] = []
    for fingerprint, number in signers.items():
        ret.append(
            ngettext("{0} commit by author {1} with key {2} now expired",
                     "{0} commits by author {1} with key {2} now expired",
                     number).format(
                number, fingerprint_to_authors[fingerprint], fingerprint))
    return ret


def verbose_valid_message(result) -> List[str]:
    """takes a verify result and returns list of signed commits strings"""
    signers: Dict[str, int] = {}
    for rev_id, validity, uid in result:
        if validity == SIGNATURE_VALID:
            signers.setdefault(uid, 0)
            signers[uid] += 1
    ret: List[str] = []
    for uid, number in signers.items():
        ret.append(ngettext("{0} signed {1} commit",
                            "{0} signed {1} commits",
                            number).format(uid, number))
    return ret


def verbose_not_valid_message(result, repo) -> List[str]:
    """takes a verify result and returns list of not valid commit info"""
    signers: Dict[str, int] = {}
    for rev_id, validity, empty in result:
        if validity == SIGNATURE_NOT_VALID:
            revision = repo.get_revision(rev_id)
            authors = ', '.join(revision.get_apparent_authors())
            signers.setdefault(authors, 0)
            signers[authors] += 1
    ret: List[str] = []
    for authors, number in signers.items():
        ret.append(ngettext("{0} commit by author {1}",
                            "{0} commits by author {1}",
                            number).format(number, authors))
    return ret


def verbose_not_signed_message(result, repo) -> List[str]:
    """takes a verify result and returns list of not signed commit info"""
    signers: Dict[str, int] = {}
    for rev_id, validity, empty in result:
        if validity == SIGNATURE_NOT_SIGNED:
            revision = repo.get_revision(rev_id)
            authors = ', '.join(revision.get_apparent_authors())
            signers.setdefault(authors, 0)
            signers[authors] += 1
    ret: List[str] = []
    for authors, number in signers.items():
        ret.append(ngettext("{0} commit by author {1}",
                            "{0} commits by author {1}",
                            number).format(number, authors))
    return ret


def verbose_missing_key_message(result) -> List[str]:
    """takes a verify result and returns list of missing key info"""
    signers: Dict[str, int] = {}
    for rev_id, validity, fingerprint in result:
        if validity == SIGNATURE_KEY_MISSING:
            signers.setdefault(fingerprint, 0)
            signers[fingerprint] += 1
    ret: List[str] = []
    for fingerprint, number in list(signers.items()):
        ret.append(ngettext("Unknown key {0} signed {1} commit",
                            "Unknown key {0} signed {1} commits",
                            number).format(fingerprint, number))
    return ret



# Copyright (C) 2005, 2006 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy.tests import TestCaseWithTransport


class TestAncestry(TestCaseWithTransport):

    def _build_branches(self):
        a_wt = self.make_branch_and_tree('A')
        self.build_tree_contents([('A/foo', b'1111\n')])
        a_wt.add('foo')
        a_wt.commit('added foo', rev_id=b'A1')

        b_wt = a_wt.controldir.sprout('B').open_workingtree()
        self.build_tree_contents([('B/foo', b'1111\n22\n')])
        b_wt.commit('modified B/foo', rev_id=b'B1')

        self.build_tree_contents([('A/foo', b'000\n1111\n')])
        a_wt.commit('modified A/foo', rev_id=b'A2')

        a_wt.merge_from_branch(b_wt.branch, b_wt.last_revision(),
                               b_wt.branch.get_rev_id(1))
        a_wt.commit('merged B into A', rev_id=b'A3')
        return a_wt, b_wt

    def _check_ancestry(self, location='', result=None):
        out = self.run_bzr(['ancestry', location])[0]
        if result is not None:
            self.assertEqualDiff(result, out)
        else:
            # A2 and B1 can be in either order, because they are parallel, and
            # thus their topological order is not defined
            result = "A1\nB1\nA2\nA3\n"
            if result != out:
                result = "A1\nA2\nB1\nA3\n"
            self.assertEqualDiff(result, out)

    def test_ancestry(self):
        """Tests 'ancestry' command"""
        self._build_branches()
        os.chdir('A')
        self._check_ancestry()

    def test_ancestry_with_location(self):
        """Tests 'ancestry' command with a specified location."""
        self._build_branches()
        self._check_ancestry('A')

    def test_ancestry_with_repo_branch(self):
        """Tests 'ancestry' command with a location that is a
        repository branch."""
        a_tree = self._build_branches()[0]

        self.make_repository('repo', shared=True)

        a_tree.controldir.sprout('repo/A')
        self._check_ancestry('repo/A')

    def test_ancestry_with_checkout(self):
        """Tests 'ancestry' command with a location that is a
        checkout of a repository branch."""
        a_tree = self._build_branches()[0]
        self.make_repository('repo', shared=True)
        repo_branch = a_tree.controldir.sprout('repo/A').open_branch()
        repo_branch.create_checkout('A-checkout')
        self._check_ancestry('A-checkout')

    def test_ancestry_with_lightweight_checkout(self):
        """Tests 'ancestry' command with a location that is a
        lightweight checkout of a repository branch."""
        a_tree = self._build_branches()[0]
        self.make_repository('repo', shared=True)
        repo_branch = a_tree.controldir.sprout('repo/A').open_branch()
        repo_branch.create_checkout('A-checkout', lightweight=True)
        self._check_ancestry('A-checkout')

    def test_ancestry_with_truncated_checkout(self):
        """Tests 'ancestry' command with a location that is a
        checkout of a repository branch with a shortened revision history."""
        a_tree = self._build_branches()[0]
        self.make_repository('repo', shared=True)
        repo_branch = a_tree.controldir.sprout('repo/A').open_branch()
        repo_branch.create_checkout('A-checkout',
                                    revision_id=repo_branch.get_rev_id(2))
        self._check_ancestry('A-checkout', "A1\nA2\n")

    def test_ancestry_with_truncated_lightweight_checkout(self):
        """Tests 'ancestry' command with a location that is a lightweight
        checkout of a repository branch with a shortened revision history."""
        a_tree = self._build_branches()[0]
        self.make_repository('repo', shared=True)
        repo_branch = a_tree.controldir.sprout('repo/A').open_branch()
        repo_branch.create_checkout('A-checkout',
                                    revision_id=repo_branch.get_rev_id(2),
                                    lightweight=True)
        self._check_ancestry('A-checkout', "A1\nA2\n")



# Copyright (C) 2008, 2009, 2010, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for content filtering conformance"""

import os

from breezy.controldir import ControlDir
from breezy.filters import ContentFilter
from breezy.switch import switch
from breezy.tests.per_workingtree import TestCaseWithWorkingTree
from breezy.workingtree import WorkingTree


def _converter_helper(chunks, fn):
    result = []
    for chunk in chunks:
        result.append(getattr(chunk, fn)())
    return iter(result)


def _swapcase(chunks, context=None):
    """A converter that swaps the case of text."""
    return _converter_helper(chunks, 'swapcase')


def _uppercase(chunks, context=None):
    """A converter that converts text to uppercase."""
    return _converter_helper(chunks, 'upper')


def _lowercase(chunks, context=None):
    """A converter that converts text to lowercase."""
    return _converter_helper(chunks, 'lower')


_trailer_string = b'\nend string\n'


def _append_text(chunks, context=None):
    """A content filter that appends a string to the end of the file.

    This tests filters that change the length."""
    return chunks + [_trailer_string]


def _remove_appended_text(chunks, context=None):
    """Remove the appended text."""

    text = b''.join(chunks)
    if text.endswith(_trailer_string):
        text = text[:-len(_trailer_string)]
    return [text]


class TestWorkingTreeWithContentFilters(TestCaseWithWorkingTree):

    def create_cf_tree(self, txt_reader, txt_writer, dir='.'):
        tree = self.make_branch_and_tree(dir)

        def _content_filter_stack(path=None, file_id=None):
            if path.endswith('.txt'):
                return [ContentFilter(txt_reader, txt_writer)]
            else:
                return []
        tree._content_filter_stack = _content_filter_stack
        self.build_tree_contents([
            (dir + '/file1.txt', b'Foo Txt'),
            (dir + '/file2.bin', b'Foo Bin')])
        tree.add(['file1.txt', 'file2.bin'])
        tree.commit('commit raw content')
        return tree, 'file1.txt', 'file2.bin'

    def create_cf_tree_with_two_revisions(self, txt_reader, txt_writer,
                                          dir='.'):
        tree = self.make_branch_and_tree(dir)

        def _content_filter_stack(path=None, file_id=None):
            if path.endswith('.txt'):
                return [ContentFilter(txt_reader, txt_writer)]
            else:
                return []
        tree._content_filter_stack = _content_filter_stack
        self.build_tree_contents([
            (dir + '/file1.txt', b'Foo Txt'),
            (dir + '/file2.bin', b'Foo Bin'),
            (dir + '/file3.txt', b'Bar Txt'),
            ])
        tree.add(['file1.txt', 'file2.bin', 'file3.txt'])
        tree.commit('commit raw content')
        # Commit another revision with various changes. We make sure
        # the change includes a modification, an addition and a deletion.
        # Renames are more complex and need a separate set of tests later.
        self.build_tree_contents([
            (dir + '/file1.txt', b'Foo ROCKS!'),
            (dir + '/file4.txt', b'Hello World'),
            ])
        tree.add(['file4.txt'])
        tree.remove(['file3.txt'], keep_files=False)
        tree.commit("change, add and rename stuff")
        return tree, 'file1.txt', 'file2.bin', 'file3.txt', 'file4.txt'

    def patch_in_content_filter(self):
        # Patch in a custom, symmetric content filter stack. It's pretty gross
        # that we need to monkey-patch a class method to do this, bit it's
        # the easiest way currently given we don't have easy access to the
        # WorkingTree after it is created but before the filter stack is used
        # to populate content.
        def new_stack(tree, path=None, file_id=None):
            if path.endswith('.txt'):
                return [ContentFilter(_swapcase, _swapcase)]
            else:
                return []
        self.overrideAttr(WorkingTree, '_content_filter_stack', new_stack)

    def assert_basis_content(self, expected_content, branch, path):
        # Note: We need to use try/finally here instead of addCleanup()
        # as the latter leaves the read lock in place too long
        basis = branch.basis_tree()
        with basis.lock_read():
            self.assertEqual(expected_content, basis.get_file_text(path))

    def test_symmetric_content_filtering(self):
        # test handling when read then write gives back the initial content
        tree, txt_path, bin_path = self.create_cf_tree(
            txt_reader=_swapcase, txt_writer=_swapcase)
        # Check that the basis tree has the expected content
        basis = tree.basis_tree()
        basis.lock_read()
        self.addCleanup(basis.unlock)
        if tree.supports_content_filtering():
            expected = b"fOO tXT"
        else:
            expected = b"Foo Txt"
        self.assertEqual(
            expected,
            basis.get_file_text(txt_path))
        self.assertEqual(
            b'Foo Bin',
            basis.get_file_text(bin_path))
        # Check that the working tree has the original content
        tree.lock_read()
        self.addCleanup(tree.unlock)
        with tree.get_file(txt_path, filtered=False) as f:
            self.assertEqual(b'Foo Txt', f.read())
        with tree.get_file(bin_path, filtered=False) as f:
            self.assertEqual(b'Foo Bin', f.read())

    def test_readonly_content_filtering(self):
        # test handling with a read filter but no write filter
        tree, txt_path, bin_path = self.create_cf_tree(
            txt_reader=_uppercase, txt_writer=None)
        # Check that the basis tree has the expected content
        basis = tree.basis_tree()
        basis.lock_read()
        self.addCleanup(basis.unlock)
        if tree.supports_content_filtering():
            expected = b"FOO TXT"
        else:
            expected = b"Foo Txt"
        self.assertEqual(expected, basis.get_file_text(txt_path))
        self.assertEqual(b'Foo Bin', basis.get_file_text(bin_path))
        # We expect the workingtree content to be unchanged (for now at least)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        with tree.get_file(txt_path, filtered=False) as f:
            self.assertEqual(b'Foo Txt', f.read())
        with tree.get_file(bin_path, filtered=False) as f:
            self.assertEqual(b'Foo Bin', f.read())

    def test_branch_source_filtered_target_not(self):
        # Create a source branch with content filtering
        source, txt_path, bin_path = self.create_cf_tree(
            txt_reader=_uppercase, txt_writer=_lowercase, dir='source')
        if not source.supports_content_filtering():
            return
        self.assertFileEqual(b"Foo Txt", 'source/file1.txt')
        self.assert_basis_content(b"FOO TXT", source, txt_path)

        # Now branch it
        self.run_bzr('branch source target')
        target = WorkingTree.open('target')
        # Even though the content in source and target are different
        # due to different filters, iter_changes should be clean
        self.assertFileEqual(b"FOO TXT", 'target/file1.txt')
        changes = target.changes_from(source.basis_tree())
        self.assertFalse(changes.has_changed())

    def test_branch_source_not_filtered_target_is(self):
        # Create a source branch with content filtering
        source, txt_path, bin_path = self.create_cf_tree(
            txt_reader=None, txt_writer=None, dir='source')
        if not source.supports_content_filtering():
            return
        self.assertFileEqual(b"Foo Txt", 'source/file1.txt')
        self.assert_basis_content(b"Foo Txt", source, txt_path)

        # Now patch in content filtering and branch the source
        self.patch_in_content_filter()
        self.run_bzr('branch source target')
        target = WorkingTree.open('target')
        # Even though the content in source and target are different
        # due to different filters, iter_changes should be clean
        self.assertFileEqual(b"fOO tXT", 'target/file1.txt')
        changes = target.changes_from(source.basis_tree())
        self.assertFalse(changes.has_changed())

    def test_path_content_summary(self):
        """path_content_summary should always talk about the canonical form."""
        # see https://bugs.launchpad.net/bzr/+bug/415508
        #
        # set up a tree where the canonical form has a string added to the
        # end
        source, txt_path, bin_path = self.create_cf_tree(
            txt_reader=_append_text,
            txt_writer=_remove_appended_text,
            dir='source')
        if not source.supports_content_filtering():
            return
        source.lock_read()
        self.addCleanup(source.unlock)

        expected_canonical_form = b'Foo Txt\nend string\n'
        with source.get_file(txt_path, filtered=True) as f:
            self.assertEqual(f.read(), expected_canonical_form)
        with source.get_file(txt_path, filtered=False) as f:
            self.assertEqual(f.read(), b'Foo Txt')

        # results are: kind, size, executable, sha1_or_link_target
        result = source.path_content_summary('file1.txt')

        self.assertEqual(result,
                         ('file', None, False, None))

        # we could give back the length of the canonical form, but in general
        # that will be expensive to compute, so it's acceptable to just return
        # None.

    def test_content_filtering_applied_on_pull(self):
        # Create a source branch with two revisions
        source, path1, path2, path3, path4 = \
            self.create_cf_tree_with_two_revisions(txt_reader=None,
                                                   txt_writer=None, dir='source')
        if not source.supports_content_filtering():
            return
        self.assertFileEqual(b"Foo ROCKS!", 'source/file1.txt')
        self.assert_basis_content(b"Foo ROCKS!", source, path1)

        # Now patch in content filtering and branch from revision 1
        self.patch_in_content_filter()
        self.run_bzr('branch -r1 source target')
        target = WorkingTree.open('target')
        self.assert_basis_content(b"Foo Txt", target, path1)
        self.assertFileEqual(b"fOO tXT", 'target/file1.txt')
        self.assert_basis_content(b"Foo Bin", target, path2)
        self.assertFileEqual(b"Foo Bin", 'target/file2.bin')
        self.assert_basis_content(b"Bar Txt", target, path3)
        self.assertFileEqual(b"bAR tXT", 'target/file3.txt')

        # Pull the latter change and check the target tree is updated
        self.run_bzr('pull -d target')
        self.assert_basis_content(b"Foo ROCKS!", target, path1)
        self.assertFileEqual(b"fOO rocks!", 'target/file1.txt')
        self.assert_basis_content(b"Foo Bin", target, path2)
        self.assert_basis_content(b"Hello World", target, path4)
        self.assertFileEqual(b"hELLO wORLD", 'target/file4.txt')

    def test_content_filtering_applied_on_merge(self):
        # Create a source branch with two revisions
        source, path1, path2, path3, path4 = \
            self.create_cf_tree_with_two_revisions(txt_reader=None,
                                                   txt_writer=None, dir='source')
        if not source.supports_content_filtering():
            return
        self.assert_basis_content(b"Foo ROCKS!", source, path1)
        self.assertFileEqual(b"Foo ROCKS!", 'source/file1.txt')
        self.assert_basis_content(b"Foo Bin", source, path2)
        self.assert_basis_content(b"Hello World", source, path4)
        self.assertFileEqual(b"Hello World", 'source/file4.txt')

        # Now patch in content filtering and branch from revision 1
        self.patch_in_content_filter()
        self.run_bzr('branch -r1 source target')
        target = WorkingTree.open('target')
        self.assert_basis_content(b"Foo Txt", target, path1)
        self.assertFileEqual(b"fOO tXT", 'target/file1.txt')
        self.assertFileEqual(b"Foo Bin", 'target/file2.bin')
        self.assertFileEqual(b"bAR tXT", 'target/file3.txt')

        # Merge the latter change and check the target tree is updated
        self.run_bzr('merge -d target source')
        self.assertFileEqual(b"fOO rocks!", 'target/file1.txt')
        self.assertFileEqual(b"hELLO wORLD", 'target/file4.txt')

        # Commit the merge and check the right content is stored
        target.commit("merge file1.txt changes from source")
        self.assert_basis_content(b"Foo ROCKS!", target, path1)
        self.assert_basis_content(b"Hello World", target, path4)

    def test_content_filtering_applied_on_switch(self):
        # Create a source branch with two revisions
        source, path1, path2, path3, path4 = \
            self.create_cf_tree_with_two_revisions(txt_reader=None,
                                                   txt_writer=None, dir='branch-a')
        if not source.supports_content_filtering():
            return

        # Now patch in content filtering and branch from revision 1
        self.patch_in_content_filter()
        self.run_bzr('branch -r1 branch-a branch-b')

        # Now create a lightweight checkout referring to branch-b
        self.run_bzr('checkout --lightweight branch-b checkout')
        self.assertFileEqual(b"fOO tXT", 'checkout/file1.txt')

        # Switch it to branch-b and check the tree is updated
        checkout_control_dir = ControlDir.open_containing('checkout')[0]
        switch(checkout_control_dir, source.branch)
        self.assertFileEqual(b"fOO rocks!", 'checkout/file1.txt')
        self.assertFileEqual(b"hELLO wORLD", 'checkout/file4.txt')

    def test_content_filtering_applied_on_revert_delete(self):
        # Create a source branch with content filtering
        source, txt_path, bin_path = self.create_cf_tree(
            txt_reader=_uppercase, txt_writer=_lowercase, dir='source')
        if not source.supports_content_filtering():
            return
        self.assertFileEqual(b"Foo Txt", 'source/file1.txt')
        self.assert_basis_content(b"FOO TXT", source, txt_path)

        # Now delete the file, revert it and check the content
        os.unlink('source/file1.txt')
        self.assertFalse(os.path.exists('source/file1.txt'))
        source.revert(['file1.txt'])
        self.assertTrue(os.path.exists('source/file1.txt'))
        # Note: we don't get back exactly what was in the tree
        # previously because lower(upper(text)) is a lossy transformation
        self.assertFileEqual(b"foo txt", 'source/file1.txt')

    def test_content_filtering_applied_on_revert_rename(self):
        # Create a source branch with content filtering
        source, txt_path, bin_path = self.create_cf_tree(
            txt_reader=_uppercase, txt_writer=_lowercase, dir='source')
        if not source.supports_content_filtering():
            return
        self.assertFileEqual(b"Foo Txt", 'source/file1.txt')
        self.assert_basis_content(b"FOO TXT", source, txt_path)

        # Now modify & rename a file, revert it and check the content
        self.build_tree_contents([
            ('source/file1.txt', b'Foo Txt with new content')])
        source.rename_one('file1.txt', 'file1.bin')
        self.assertTrue(os.path.exists('source/file1.bin'))
        self.assertFalse(os.path.exists('source/file1.txt'))
        self.assertFileEqual(b"Foo Txt with new content", 'source/file1.bin')
        source.revert(['file1.bin'])
        self.assertFalse(os.path.exists('source/file1.bin'))
        self.assertTrue(os.path.exists('source/file1.txt'))
        # Note: we don't get back exactly what was in the tree
        # previously because lower(upper(text)) is a lossy transformation
        self.assertFileEqual(b"foo txt", 'source/file1.txt')



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import sys
import tempfile

import breezy

from .. import controldir, errors, merge_directive, osutils
from ..bzr import generate_ids
from ..bzr.conflicts import ContentsConflict, PathConflict, TextConflict
from ..merge import Diff3Merger, Merge3Merger, Merger, WeaveMerger
from ..osutils import getcwd, pathjoin
from ..workingtree import WorkingTree
from . import TestCaseWithTransport, TestSkipped, features


class MergeBuilder:

    def __init__(self, dir=None):
        self.dir = tempfile.mkdtemp(prefix="merge-test", dir=dir)
        self.tree_root = generate_ids.gen_root_id()

        def wt(name):
            path = pathjoin(self.dir, name)
            os.mkdir(path)
            wt = controldir.ControlDir.create_standalone_workingtree(path)
            # the tests perform pulls, so need a branch that is writeable.
            wt.lock_write()
            if wt.supports_file_ids:
                wt.set_root_id(self.tree_root)
            wt.flush()
            tt = wt.transform()
            return wt, tt
        self.base, self.base_tt = wt('base')
        self.this, self.this_tt = wt('this')
        self.other, self.other_tt = wt('other')

    def root(self):
        ret = []
        for tt in [self.this_tt, self.base_tt, self.other_tt]:
            ret.append(tt.trans_id_tree_path(''))
        return tuple(ret)

    def get_cset_path(self, parent, name):
        if name is None:
            if parent is not None:
                raise AssertionError()
            return None
        return pathjoin(self.cset.entries[parent].path, name)

    def add_file(self, parent, name, contents, executable, this=True,
                 base=True, other=True, file_id=None):
        ret = []
        for i, (option, tt) in enumerate(self.selected_transforms(this, base, other)):
            if option is True:
                trans_id = tt.new_file(name, parent[i], [contents], executable=executable, file_id=file_id)
            else:
                trans_id = None
            ret.append(trans_id)
        return tuple(ret)

    def merge(self, merge_type=Merge3Merger, interesting_files=None, **kwargs):
        merger = self.make_merger(merge_type, interesting_files, **kwargs)
        merger.do_merge()
        return merger.cooked_conflicts

    def make_preview_transform(self):
        merger = self.make_merger(Merge3Merger, None, this_revision_tree=True)
        return merger.make_preview_transform()

    def make_merger(self, merge_type, interesting_files,
                    this_revision_tree=False, **kwargs):
        self.base_tt.apply()
        self.base.commit('base commit')
        for tt, wt in ((self.this_tt, self.this), (self.other_tt, self.other)):
            # why does this not do wt.pull() ?
            wt.branch.pull(self.base.branch)
            wt.set_parent_ids([wt.branch.last_revision()])
            wt.flush()
            # We maintain a write lock, so make sure changes are flushed to
            # disk first
            tt.apply()
            wt.commit('branch commit')
            wt.flush()
            if wt.branch.last_revision_info()[0] != 2:
                raise AssertionError()
        self.this.branch.fetch(self.other.branch)
        other_basis = self.other.branch.basis_tree()
        if this_revision_tree:
            self.this.commit('message')
            this_tree = self.this.basis_tree()
        else:
            this_tree = self.this
        merger = merge_type(this_tree, self.this, self.base, other_basis,
                            interesting_files=interesting_files, do_merge=False,
                            this_branch=self.this.branch, **kwargs)
        return merger

    def list_transforms(self):
        return [self.this_tt, self.base_tt, self.other_tt]

    def selected_transforms(self, this, base, other):
        pairs = [(this, self.this_tt), (base, self.base_tt),
                 (other, self.other_tt)]
        return [(v, tt) for (v, tt) in pairs if v is not None]

    def add_symlink(self, parent, name, contents, file_id=None):
        ret = []
        for i, tt in enumerate(self.list_transforms()):
            trans_id = tt.new_symlink(name, parent[i], contents, file_id=file_id)
            ret.append(trans_id)
        return ret

    def remove_file(self, trans_ids, base=False, this=False, other=False):
        for trans_id, (option, tt) in zip(trans_ids, self.selected_transforms(this, base, other)):
            if option is True:
                tt.cancel_creation(trans_id)
                tt.cancel_versioning(trans_id)
                tt.set_executability(None, trans_id)

    def add_dir(self, parent, name, this=True, base=True, other=True, file_id=None):
        ret = []
        for i, (option, tt) in enumerate(self.selected_transforms(this, base, other)):
            if option is True:
                trans_id = tt.new_directory(name, parent[i], file_id)
            else:
                trans_id = None
            ret.append(trans_id)
        return tuple(ret)

    def change_name(self, trans_ids, base=None, this=None, other=None):
        for val, tt, trans_id in (
                        (base, self.base_tt, trans_ids[0]),
                        (this, self.this_tt, trans_ids[1]),
                        (other, self.other_tt, trans_ids[2])):
            if val is None:
                continue
            parent_id = tt.final_parent(trans_id)
            tt.adjust_path(val, parent_id, trans_id)

    def change_parent(self, trans_ids, base=None, this=None, other=None):
        for trans_id, (parent, tt) in zip(trans_ids, self.selected_transforms(this, base, other)):
            parent_id = tt.trans_id_file_id(parent)
            tt.adjust_path(tt.final_name(trans_id), parent_id, trans_id)

    def change_contents(self, trans_id, base=None, this=None, other=None):
        for trans_id, (contents, tt) in zip(trans_id, self.selected_transforms(this, base, other)):
            tt.cancel_creation(trans_id)
            tt.create_file([contents], trans_id)

    def change_target(self, trans_ids, base=None, this=None, other=None):
        for trans_id, (target, tt) in zip(trans_ids, self.selected_transforms(this, base, other)):
            tt.cancel_creation(trans_id)
            tt.create_symlink(target, trans_id)

    def change_perms(self, trans_ids, base=None, this=None, other=None):
        for trans_id, (executability, tt) in zip(trans_ids, self.selected_transforms(this, base, other)):
            tt.set_executability(None, trans_id)
            tt.set_executability(executability, trans_id)

    def apply_inv_change(self, inventory_change, orig_inventory):
        orig_inventory_by_path = {}
        for file_id, path in orig_inventory.items():
            orig_inventory_by_path[path] = file_id

        def parent_id(file_id):
            try:
                parent_dir = os.path.dirname(orig_inventory[file_id])
            except:
                print(file_id)
                raise
            if parent_dir == "":
                return None
            return orig_inventory_by_path[parent_dir]

        def new_path(file_id):
            if fild_id in inventory_change:
                return inventory_change[file_id]
            else:
                parent = parent_id(file_id)
                if parent is None:
                    return orig_inventory[file_id]
                dirname = new_path(parent)
                return pathjoin(dirname, os.path.basename(orig_inventory[file_id]))

        new_inventory = {}
        for file_id in orig_inventory:
            path = new_path(file_id)
            if path is None:
                continue
            new_inventory[file_id] = path

        for file_id, path in inventory_change.items():
            if file_id in orig_inventory:
                continue
            new_inventory[file_id] = path
        return new_inventory

    def unlock(self):
        self.base.unlock()
        self.this.unlock()
        self.other.unlock()

    def cleanup(self):
        self.unlock()
        osutils.rmtree(self.dir)


class MergeTest(TestCaseWithTransport):

    def test_change_name(self):
        """Test renames"""
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"hello1", True, file_id=b"1")
        builder.change_name(name1, other="name2")
        name3 = builder.add_file(builder.root(), "name3", b"hello2", True, file_id=b"2")
        builder.change_name(name3, base="name4")
        name5 = builder.add_file(builder.root(), "name5", b"hello3", True, file_id=b"3")
        builder.change_name(name5, this="name6")
        builder.merge()
        builder.cleanup()
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"hello1", False, file_id=b"1")
        builder.change_name(name1, other="name2", this="name3")
        conflicts = builder.merge()
        self.assertEqual(conflicts, [PathConflict('name3', 'name2', b'1')])
        builder.cleanup()

    def test_merge_one(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"hello1", True, file_id=b'1')
        builder.change_contents(name1, other=b"text4")
        name2 = builder.add_file(builder.root(), "name2", b"hello1", True, file_id=b"2")
        builder.change_contents(name2, other=b"text4")
        builder.merge(interesting_files=["name1"])
        self.assertEqual(builder.this.get_file("name1").read(), b"text4")
        self.assertEqual(builder.this.get_file("name2").read(), b"hello1")
        builder.cleanup()

    def test_file_moves(self):
        """Test moves"""
        builder = MergeBuilder(getcwd())
        dir1 = builder.add_dir(builder.root(), "dir1", file_id=b"1")
        dir2 = builder.add_dir(builder.root(), "dir2", file_id=b"2")
        file1 = builder.add_file(dir1, "file1", b"hello1", True, file_id=b"3")
        file2 = builder.add_file(dir1, "file2", b"hello2", True, file_id=b'4')
        file3 = builder.add_file(dir1, "file3", b"hello3", True, file_id=b'5')
        builder.change_parent(file1, other=b"2")
        builder.change_parent(file2, this=b"2")
        builder.change_parent(file3, base=b"2")
        builder.merge()
        builder.cleanup()

        builder = MergeBuilder(getcwd())
        dir1 = builder.add_dir(builder.root(), "dir1", file_id=b'1')
        builder.add_dir(builder.root(), "dir2", file_id=b'2')
        builder.add_dir(builder.root(), "dir3", file_id=b'3')
        file1 = builder.add_file(dir1, "file1", b"hello1", False, file_id=b"4")
        builder.change_parent(file1, other=b"2", this=b"3")
        conflicts = builder.merge()
        path2 = pathjoin('dir2', 'file1')
        path3 = pathjoin('dir3', 'file1')
        self.assertEqual(conflicts, [PathConflict(path3, path2, b'4')])
        builder.cleanup()

    def test_contents_merge(self):
        """Test merge3 merging"""
        self.do_contents_test(Merge3Merger)

    def test_contents_merge2(self):
        """Test diff3 merging"""
        if sys.platform == 'win32':
            raise TestSkipped("diff3 does not have --binary flag"
                              " and therefore always fails on win32")
        try:
            self.do_contents_test(Diff3Merger)
        except errors.NoDiff3:
            raise TestSkipped("diff3 not available")

    def test_contents_merge3(self):
        """Test diff3 merging"""
        self.do_contents_test(WeaveMerger)

    def test_reprocess_weave(self):
        # Reprocess works on weaves, and behaves as expected
        builder = MergeBuilder(getcwd())
        blah = builder.add_file(builder.root(), 'blah', b'a', False, file_id=b'a')
        builder.change_contents(
            blah, this=b'b\nc\nd\ne\n', other=b'z\nc\nd\ny\n')
        builder.merge(WeaveMerger, reprocess=True)
        expected = b"""<<<<<<< TREE
b
=======
z
>>>>>>> MERGE-SOURCE
c
d
<<<<<<< TREE
e
=======
y
>>>>>>> MERGE-SOURCE
"""
        self.assertEqualDiff(
            builder.this.get_file_text('blah'),
            expected)
        builder.cleanup()

    def do_contents_test(self, merge_factory):
        """Test merging with specified ContentsChange factory"""
        builder = self.contents_test_success(merge_factory)
        builder.cleanup()
        self.contents_test_conflicts(merge_factory)

    def contents_test_success(self, merge_factory):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_contents(name1, other=b"text4")
        name3 = builder.add_file(builder.root(), "name3", b"text2", False, file_id=b"2")
        builder.change_contents(name3, base=b"text5")
        builder.add_file(builder.root(), "name5", b"text3", True, file_id=b"3")
        name6 = builder.add_file(builder.root(), "name6", b"text4", True, file_id=b"4")
        builder.remove_file(name6, base=True)
        name7 = builder.add_file(builder.root(), "name7", b"a\nb\nc\nd\ne\nf\n",
                         True, file_id=b"5")
        builder.change_contents(name7, other=b"a\nz\nc\nd\ne\nf\n",
                                this=b"a\nb\nc\nd\ne\nz\n")
        conflicts = builder.merge(merge_factory)
        try:
            self.assertEqual([], conflicts)
            self.assertEqual(b"text4", builder.this.get_file("name1").read())
            self.assertEqual(b"text2", builder.this.get_file("name3").read())
            self.assertEqual(b"a\nz\nc\nd\ne\nz\n",
                             builder.this.get_file("name7").read())
            self.assertTrue(builder.this.is_executable("name1"))
            self.assertFalse(builder.this.is_executable("name3"))
            self.assertTrue(builder.this.is_executable("name5"))
        except:
            builder.unlock()
            raise
        return builder

    def contents_test_conflicts(self, merge_factory):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_contents(name1, other=b"text4", this=b"text3")
        name2 = builder.add_file(builder.root(), "name2", b"text1", True, file_id=b"2")
        builder.change_contents(name2, other=b"\x00", this=b"text3")
        name3 = builder.add_file(builder.root(), "name3", b"text5", False, file_id=b"3")
        builder.change_perms(name3, this=True)
        builder.change_contents(name3, this=b'moretext')
        builder.remove_file(name3, other=True)
        conflicts = builder.merge(merge_factory)
        self.assertEqual(conflicts, [TextConflict('name1', file_id=b'1'),
                                     ContentsConflict('name2', file_id=b'2'),
                                     ContentsConflict('name3', file_id=b'3')])
        with builder.this.get_file(builder.this.id2path(b'2')) as f:
            self.assertEqual(f.read(), b'\x00')
        builder.cleanup()

    def test_symlink_conflicts(self):
        if sys.platform != "win32":
            builder = MergeBuilder(getcwd())
            name2 = builder.add_symlink(builder.root(), "name2", "target1", file_id=b"2")
            builder.change_target(name2, other="target4", base="text3")
            conflicts = builder.merge()
            self.assertEqual(conflicts, [ContentsConflict('name2',
                                                          file_id=b'2')])
            builder.cleanup()

    def test_symlink_merge(self):
        if sys.platform != "win32":
            builder = MergeBuilder(getcwd())
            name1 = builder.add_symlink(builder.root(), "name1", "target1", file_id=b"1")
            name2 = builder.add_symlink(builder.root(), "name2", "target1", file_id=b"2")
            name3 = builder.add_symlink(builder.root(), "name3", "target1", file_id=b'3')
            builder.change_target(name1, this=b"target2")
            builder.change_target(name2, base=b"target2")
            builder.change_target(name3, other=b"target2")
            builder.merge()
            self.assertEqual(
                builder.this.get_symlink_target("name1"), "target2")
            self.assertEqual(
                builder.this.get_symlink_target("name2"), "target1")
            self.assertEqual(
                builder.this.get_symlink_target("name3"), "target2")
            builder.cleanup()

    def test_no_passive_add(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.remove_file(name1, this=True)
        builder.merge()
        builder.cleanup()

    def test_perms_merge(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_perms(name1, other=False)
        name2 = builder.add_file(builder.root(), "name2", b"text2", True, file_id=b"2")
        builder.change_perms(name2, base=False)
        name3 = builder.add_file(builder.root(), "name3", b"text3", True, file_id=b"3")
        builder.change_perms(name3, this=False)
        name4 = builder.add_file(builder.root(), 'name4', b'text4', False, file_id=b'4')
        builder.change_perms(name4, this=True)
        builder.remove_file(name4, base=True)
        builder.merge()
        self.assertIs(builder.this.is_executable("name1"), False)
        self.assertIs(builder.this.is_executable("name2"), True)
        self.assertIs(builder.this.is_executable("name3"), False)
        builder.cleanup()

    def test_new_suffix(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", True, file_id=b"1")
        builder.change_contents(name1, other=b"text3")
        builder.add_file(builder.root(), "name1.new", b"text2", True, file_id=b"2")
        builder.merge()
        os.lstat(builder.this.abspath('name1.new'))
        builder.cleanup()

    def test_spurious_conflict(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), "name1", b"text1", False, file_id=b"1")
        builder.remove_file(name1, other=True)
        builder.add_file(builder.root(), "name1", b"text1", False,
                         this=False, base=False, file_id=b"2")
        conflicts = builder.merge()
        self.assertEqual(conflicts, [])
        builder.cleanup()

    def test_merge_one_renamed(self):
        builder = MergeBuilder(getcwd())
        name1 = builder.add_file(builder.root(), 'name1', b'text1a', False, file_id=b'1')
        builder.change_name(name1, this='name2')
        builder.change_contents(name1, other=b'text2')
        builder.merge(interesting_files=['name2'])
        self.assertEqual(b'text2', builder.this.get_file('name2').read())
        builder.cleanup()


class FunctionalMergeTest(TestCaseWithTransport):

    def test_trivial_star_merge(self):
        """Test that merges in a star shape Just Work."""
        # John starts a branch
        self.build_tree(("original/", "original/file1", "original/file2"))
        tree = self.make_branch_and_tree('original')
        branch = tree.branch
        tree.smart_add(["original"])
        tree.commit("start branch.", verbose=False)
        # Mary branches it.
        self.build_tree(("mary/",))
        branch.controldir.clone("mary")
        # Now John commits a change
        with open("original/file1", "w") as f:
            f.write("John\n")
        tree.commit("change file1")
        # Mary does too
        mary_tree = WorkingTree.open('mary')
        mary_branch = mary_tree.branch
        with open("mary/file2", "w") as f:
            f.write("Mary\n")
        mary_tree.commit("change file2")
        # john should be able to merge with no conflicts.
        base = [None, None]
        other = ("mary", -1)
        tree.merge_from_branch(mary_tree.branch)
        with open("original/file1") as f:
            self.assertEqual("John\n", f.read())
        with open("original/file2") as f:
            self.assertEqual("Mary\n", f.read())

    def test_conflicts(self):
        wta = self.make_branch_and_tree('a')
        self.build_tree_contents([('a/file', b'contents\n')])
        wta.add('file')
        wta.commit('base revision', allow_pointless=False)
        d_b = wta.branch.controldir.clone('b')
        self.build_tree_contents([('a/file', b'other contents\n')])
        wta.commit('other revision', allow_pointless=False)
        self.build_tree_contents([('b/file', b'this contents contents\n')])
        wtb = d_b.open_workingtree()
        wtb.commit('this revision', allow_pointless=False)
        self.assertEqual(1, len(wtb.merge_from_branch(wta.branch)))
        self.assertPathExists('b/file.THIS')
        self.assertPathExists('b/file.BASE')
        self.assertPathExists('b/file.OTHER')
        wtb.revert()
        self.assertEqual(1, len(wtb.merge_from_branch(wta.branch, merge_type=WeaveMerger)))
        self.assertPathExists('b/file')
        self.assertPathExists('b/file.THIS')
        self.assertPathExists('b/file.BASE')
        self.assertPathExists('b/file.OTHER')

    def test_weave_conflicts_not_in_base(self):
        builder = self.make_branch_builder('source')
        builder.start_series()
        # See bug #494197
        #  A        base revision (before criss-cross)
        #  |\
        #  B C      B does nothing, C adds 'foo'
        #  |X|
        #  D E      D and E modify foo in incompatible ways
        #
        # Merging will conflict, with C as a clean base text. However, the
        # current code uses A as the global base and 'foo' doesn't exist there.
        # It isn't trivial to create foo.BASE because it tries to look up
        # attributes like 'executable' in A.
        a_id = builder.build_snapshot(None, [
            ('add', ('', b'TREE_ROOT', 'directory', None))])
        b_id = builder.build_snapshot([a_id], [])
        c_id = builder.build_snapshot([a_id], [
            ('add', ('foo', b'foo-id', 'file', b'orig\ncontents\n'))])
        d_id = builder.build_snapshot([b_id, c_id], [
            ('add', ('foo', b'foo-id', 'file', b'orig\ncontents\nand D\n'))])
        e_id = builder.build_snapshot([c_id, b_id], [
            ('modify', ('foo', b'orig\ncontents\nand E\n'))])
        builder.finish_series()
        tree = builder.get_branch().create_checkout('tree', lightweight=True)
        self.assertEqual(1, len(tree.merge_from_branch(
            tree.branch, to_revision=d_id, merge_type=WeaveMerger)))
        self.assertPathExists('tree/foo.THIS')
        self.assertPathExists('tree/foo.OTHER')
        self.expectFailure('fail to create .BASE in some criss-cross merges',
                           self.assertPathExists, 'tree/foo.BASE')
        self.assertPathExists('tree/foo.BASE')

    def test_merge_unrelated(self):
        """Sucessfully merges unrelated branches with no common names"""
        wta = self.make_branch_and_tree('a')
        a = wta.branch
        with open('a/a_file', 'wb') as f:
            f.write(b'contents\n')
        wta.add('a_file')
        wta.commit('a_revision', allow_pointless=False)
        wtb = self.make_branch_and_tree('b')
        b = wtb.branch
        with open('b/b_file', 'wb') as f:
            f.write(b'contents\n')
        wtb.add('b_file')
        b_rev = wtb.commit('b_revision', allow_pointless=False)
        wta.merge_from_branch(wtb.branch, b_rev, b'null:')
        self.assertTrue(os.path.lexists('a/b_file'))
        self.assertEqual([b_rev], wta.get_parent_ids()[1:])

    def test_merge_unrelated_conflicting(self):
        """Sucessfully merges unrelated branches with common names"""
        wta = self.make_branch_and_tree('a')
        a = wta.branch
        with open('a/file', 'wb') as f:
            f.write(b'contents\n')
        wta.add('file')
        wta.commit('a_revision', allow_pointless=False)
        wtb = self.make_branch_and_tree('b')
        b = wtb.branch
        with open('b/file', 'wb') as f:
            f.write(b'contents\n')
        wtb.add('file')
        b_rev = wtb.commit('b_revision', allow_pointless=False)
        wta.merge_from_branch(wtb.branch, b_rev, b'null:')
        self.assertTrue(os.path.lexists('a/file'))
        self.assertTrue(os.path.lexists('a/file.moved'))
        self.assertEqual([b_rev], wta.get_parent_ids()[1:])

    def test_merge_deleted_conflicts(self):
        wta = self.make_branch_and_tree('a')
        with open('a/file', 'wb') as f:
            f.write(b'contents\n')
        wta.add('file')
        wta.commit('a_revision', allow_pointless=False)
        self.run_bzr('branch a b')
        os.remove('a/file')
        wta.commit('removed file', allow_pointless=False)
        with open('b/file', 'wb') as f:
            f.write(b'changed contents\n')
        wtb = WorkingTree.open('b')
        wtb.commit('changed file', allow_pointless=False)
        wtb.merge_from_branch(wta.branch, wta.branch.last_revision(),
                              wta.branch.get_rev_id(1))
        self.assertFalse(os.path.lexists('b/file'))

    def test_merge_metadata_vs_deletion(self):
        """Conflict deletion vs metadata change"""
        a_wt = self.make_branch_and_tree('a')
        with open('a/file', 'wb') as f:
            f.write(b'contents\n')
        a_wt.add('file')
        a_wt.commit('r0')
        self.run_bzr('branch a b')
        b_wt = WorkingTree.open('b')
        os.chmod('b/file', 0o755)
        os.remove('a/file')
        a_wt.commit('removed a')
        self.assertEqual(a_wt.branch.revno(), 2)
        self.assertFalse(os.path.exists('a/file'))
        b_wt.commit('exec a')
        a_wt.merge_from_branch(b_wt.branch, b_wt.last_revision(), b'null:')
        self.assertTrue(os.path.exists('a/file'))

    def test_merge_swapping_renames(self):
        a_wt = self.make_branch_and_tree('a')
        with open('a/un', 'wb') as f:
            f.write(b'UN')
        with open('a/deux', 'wb') as f:
            f.write(b'DEUX')
        a_wt.add('un')
        a_wt.add('deux')
        a_wt.commit('r0', rev_id=b'r0')
        self.run_bzr('branch a b')
        b_wt = WorkingTree.open('b')
        b_wt.rename_one('un', 'tmp')
        b_wt.rename_one('deux', 'un')
        b_wt.rename_one('tmp', 'deux')
        b_wt.commit('r1', rev_id=b'r1')
        self.assertEqual(
            0, len(a_wt.merge_from_branch(
                b_wt.branch, b_wt.branch.last_revision(),
                b_wt.branch.get_rev_id(1))))
        self.assertPathExists('a/un')
        self.assertTrue('a/deux')
        self.assertFalse(os.path.exists('a/tmp'))
        with open('a/un') as f:
            self.assertEqual(f.read(), 'DEUX')
        with open('a/deux') as f:
            self.assertEqual(f.read(), 'UN')

    def test_merge_delete_and_add_same(self):
        a_wt = self.make_branch_and_tree('a')
        with open('a/file', 'wb') as f:
            f.write(b'THIS')
        a_wt.add('file')
        a_wt.commit('r0')
        self.run_bzr('branch a b')
        b_wt = WorkingTree.open('b')
        os.remove('b/file')
        b_wt.commit('r1')
        with open('b/file', 'wb') as f:
            f.write(b'THAT')
        b_wt.add('file')
        b_wt.commit('r2')
        a_wt.merge_from_branch(b_wt.branch, b_wt.branch.last_revision(),
                               b_wt.branch.get_rev_id(1))
        self.assertTrue(os.path.exists('a/file'))
        with open('a/file') as f:
            self.assertEqual(f.read(), 'THAT')

    def test_merge_rename_before_create(self):
        """rename before create

        This case requires that you must not do creates
        before move-into-place:

        $ touch foo
        $ bzr add foo
        $ bzr commit
        $ bzr mv foo bar
        $ touch foo
        $ bzr add foo
        $ bzr commit
        """
        a_wt = self.make_branch_and_tree('a')
        with open('a/foo', 'wb') as f:
            f.write(b'A/FOO')
        a_wt.add('foo')
        a_wt.commit('added foo')
        self.run_bzr('branch a b')
        b_wt = WorkingTree.open('b')
        b_wt.rename_one('foo', 'bar')
        with open('b/foo', 'wb') as f:
            f.write(b'B/FOO')
        b_wt.add('foo')
        b_wt.commit('moved foo to bar, added new foo')
        a_wt.merge_from_branch(b_wt.branch, b_wt.branch.last_revision(),
                               b_wt.branch.get_rev_id(1))

    def test_merge_create_before_rename(self):
        """create before rename, target parents before children

        This case requires that you must not do move-into-place
        before creates, and that you must not do children after
        parents:

        $ touch foo
        $ bzr add foo
        $ bzr commit
        $ bzr mkdir bar
        $ bzr add bar
        $ bzr mv foo bar/foo
        $ bzr commit
        """
        os.mkdir('a')
        a_wt = self.make_branch_and_tree('a')
        with open('a/foo', 'wb') as f:
            f.write(b'A/FOO')
        a_wt.add('foo')
        a_wt.commit('added foo')
        self.run_bzr('branch a b')
        b_wt = WorkingTree.open('b')
        os.mkdir('b/bar')
        b_wt.add('bar')
        b_wt.rename_one('foo', 'bar/foo')
        b_wt.commit('created bar dir, moved foo into bar')
        a_wt.merge_from_branch(b_wt.branch, b_wt.branch.last_revision(),
                               b_wt.branch.get_rev_id(1))

    def test_merge_rename_to_temp_before_delete(self):
        """rename to temp before delete, source children before parents

        This case requires that you must not do deletes before
        move-out-of-the-way, and that you must not do children
        after parents:

        $ mkdir foo
        $ touch foo/bar
        $ bzr add foo/bar
        $ bzr commit
        $ bzr mv foo/bar bar
        $ rmdir foo
        $ bzr commit
        """
        a_wt = self.make_branch_and_tree('a')
        os.mkdir('a/foo')
        with open('a/foo/bar', 'wb') as f:
            f.write(b'A/FOO/BAR')
        a_wt.add('foo')
        a_wt.add('foo/bar')
        a_wt.commit('added foo/bar')
        self.run_bzr('branch a b')
        b_wt = WorkingTree.open('b')
        b_wt.rename_one('foo/bar', 'bar')
        os.rmdir('b/foo')
        b_wt.remove('foo')
        b_wt.commit('moved foo/bar to bar, deleted foo')
        a_wt.merge_from_branch(b_wt.branch, b_wt.branch.last_revision(),
                               b_wt.branch.get_rev_id(1))

    def test_merge_delete_before_rename_to_temp(self):
        """delete before rename to temp

        This case requires that you must not do
        move-out-of-the-way before deletes:

        $ touch foo
        $ touch bar
        $ bzr add foo bar
        $ bzr commit
        $ rm foo
        $ bzr rm foo
        $ bzr mv bar foo
        $ bzr commit
        """
        a_wt = self.make_branch_and_tree('a')
        with open('a/foo', 'wb') as f:
            f.write(b'A/FOO')
        with open('a/bar', 'wb') as f:
            f.write(b'A/BAR')
        a_wt.add('foo')
        a_wt.add('bar')
        a_wt.commit('added foo and bar')
        self.run_bzr('branch a b')
        b_wt = WorkingTree.open('b')
        os.unlink('b/foo')
        b_wt.remove('foo')
        b_wt.rename_one('bar', 'foo')
        b_wt.commit('deleted foo, renamed bar to foo')
        a_wt.merge_from_branch(b_wt.branch, b_wt.branch.last_revision(),
                               b_wt.branch.get_rev_id(1))


class TestMerger(TestCaseWithTransport):

    def set_up_trees(self):
        this = self.make_branch_and_tree('this')
        this.commit('rev1', rev_id=b'rev1')
        other = this.controldir.sprout('other').open_workingtree()
        this.commit('rev2a', rev_id=b'rev2a')
        other.commit('rev2b', rev_id=b'rev2b')
        return this, other

    def test_from_revision_ids(self):
        this, other = self.set_up_trees()
        self.assertRaises(errors.NoSuchRevision, Merger.from_revision_ids,
                          this, b'rev2b')
        this.lock_write()
        self.addCleanup(this.unlock)
        merger = Merger.from_revision_ids(this,
                                          b'rev2b', other_branch=other.branch)
        self.assertEqual(b'rev2b', merger.other_rev_id)
        self.assertEqual(b'rev1', merger.base_rev_id)
        merger = Merger.from_revision_ids(this,
                                          b'rev2b', b'rev2a', other_branch=other.branch)
        self.assertEqual(b'rev2a', merger.base_rev_id)

    def test_from_uncommitted(self):
        this, other = self.set_up_trees()
        merger = Merger.from_uncommitted(this, other, None)
        self.assertIs(other, merger.other_tree)
        self.assertIs(None, merger.other_rev_id)
        self.assertEqual(b'rev2b', merger.base_rev_id)

    def prepare_for_merging(self):
        this, other = self.set_up_trees()
        other.commit('rev3', rev_id=b'rev3')
        this.lock_write()
        self.addCleanup(this.unlock)
        return this, other

    def test_from_mergeable(self):
        this, other = self.prepare_for_merging()
        md = merge_directive.MergeDirective2.from_objects(
            other.branch.repository, b'rev3', 0, 0, 'this')
        other.lock_read()
        self.addCleanup(other.unlock)
        merger, verified = Merger.from_mergeable(this, md)
        md.patch = None
        merger, verified = Merger.from_mergeable(this, md)
        self.assertEqual('inapplicable', verified)
        self.assertEqual(b'rev3', merger.other_rev_id)
        self.assertEqual(b'rev1', merger.base_rev_id)
        md.base_revision_id = b'rev2b'
        merger, verified = Merger.from_mergeable(this, md)
        self.assertEqual(b'rev2b', merger.base_rev_id)

    def test_from_mergeable_old_merge_directive(self):
        this, other = self.prepare_for_merging()
        other.lock_write()
        self.addCleanup(other.unlock)
        md = merge_directive.MergeDirective.from_objects(
            other.branch.repository, b'rev3', 0, 0, 'this')
        merger, verified = Merger.from_mergeable(this, md)
        self.assertEqual(b'rev3', merger.other_rev_id)
        self.assertEqual(b'rev1', merger.base_rev_id)



from ...revision import Revision
from ...tests import TestCase, TestCaseWithTransport
from .cmds import collapse_by_person, get_revisions_and_committers


class TestGetRevisionsAndCommitters(TestCaseWithTransport):

    def test_simple(self):
        wt = self.make_branch_and_tree('.')
        wt.commit(message='1', committer='Fero <fero@example.com>', rev_id=b'1')
        wt.commit(message='2', committer='Fero <fero@example.com>', rev_id=b'2')
        wt.commit(message='3', committer='Jano <jano@example.com>', rev_id=b'3')
        wt.commit(message='4', committer='Jano <jano@example.com>',
                  authors=['Vinco <vinco@example.com>'], rev_id=b'4')
        wt.commit(message='5', committer='Ferko <fero@example.com>', rev_id=b'5')
        revs, committers = get_revisions_and_committers(wt.branch.repository,
                                                        [b'1', b'2', b'3', b'4', b'5'])
        fero = ('Fero', 'fero@example.com')
        jano = ('Jano', 'jano@example.com')
        vinco = ('Vinco', 'vinco@example.com')
        ferok = ('Ferko', 'fero@example.com')
        self.assertEqual({fero: fero, jano: jano, vinco: vinco, ferok: fero},
                         committers)

    def test_empty_email(self):
        wt = self.make_branch_and_tree('.')
        wt.commit(message='1', committer='Fero', rev_id=b'1')
        wt.commit(message='2', committer='Fero', rev_id=b'2')
        wt.commit(message='3', committer='Jano', rev_id=b'3')
        revs, committers = get_revisions_and_committers(wt.branch.repository,
                                                        [b'1', b'2', b'3'])
        self.assertEqual({('Fero', ''): ('Fero', ''),
                          ('Jano', ''): ('Jano', ''),
                          }, committers)

    def test_different_case(self):
        wt = self.make_branch_and_tree('.')
        wt.commit(message='1', committer='Fero', rev_id=b'1')
        wt.commit(message='2', committer='Fero', rev_id=b'2')
        wt.commit(message='3', committer='FERO', rev_id=b'3')
        revs, committers = get_revisions_and_committers(wt.branch.repository,
                                                        [b'1', b'2', b'3'])
        self.assertEqual({('Fero', ''): ('Fero', ''),
                          ('FERO', ''): ('Fero', ''),
                          }, committers)
        self.assertEqual([b'1', b'2', b'3'], sorted(
            [r.revision_id for r in revs]))


class TestCollapseByPerson(TestCase):

    def test_no_conflicts(self):
        revisions = [
            Revision('1', {}, committer='Foo <foo@example.com>'),
            Revision('2', {}, committer='Bar <bar@example.com>'),
            Revision('3', {}, committer='Bar <bar@example.com>'),
        ]
        foo = ('Foo', 'foo@example.com')
        bar = ('Bar', 'bar@example.com')
        committers = {foo: foo, bar: bar}
        info = collapse_by_person(revisions, committers)
        self.assertEqual(2, info[0][0])
        self.assertEqual({'bar@example.com': 2}, info[0][2])
        self.assertEqual({'Bar': 2}, info[0][3])

    def test_different_email(self):
        revisions = [
            Revision('1', {}, committer='Foo <foo@example.com>'),
            Revision('2', {}, committer='Foo <bar@example.com>'),
            Revision('3', {}, committer='Foo <bar@example.com>'),
        ]
        foo = ('Foo', 'foo@example.com')
        bar = ('Foo', 'bar@example.com')
        committers = {foo: foo, bar: foo}
        info = collapse_by_person(revisions, committers)
        self.assertEqual(3, info[0][0])
        self.assertEqual(
            {'foo@example.com': 1, 'bar@example.com': 2}, info[0][2])
        self.assertEqual({'Foo': 3}, info[0][3])

    def test_different_name(self):
        revisions = [
            Revision('1', {}, committer='Foo <foo@example.com>'),
            Revision('2', {}, committer='Bar <foo@example.com>'),
            Revision('3', {}, committer='Bar <foo@example.com>'),
        ]
        foo = ('Foo', 'foo@example.com')
        bar = ('Bar', 'foo@example.com')
        committers = {foo: foo, bar: foo}
        info = collapse_by_person(revisions, committers)
        self.assertEqual(3, info[0][0])
        self.assertEqual({'foo@example.com': 3}, info[0][2])
        self.assertEqual({'Foo': 1, 'Bar': 2}, info[0][3])

    def test_different_name_case(self):
        revisions = [
            Revision('1', {}, committer='Foo <foo@example.com>'),
            Revision('2', {}, committer='Foo <foo@example.com>'),
            Revision('3', {}, committer='FOO <bar@example.com>'),
        ]
        foo = ('Foo', 'foo@example.com')
        FOO = ('FOO', 'bar@example.com')
        committers = {foo: foo, FOO: foo}
        info = collapse_by_person(revisions, committers)
        self.assertEqual(3, info[0][0])
        self.assertEqual(
            {'foo@example.com': 2, 'bar@example.com': 1}, info[0][2])
        self.assertEqual({'Foo': 2, 'FOO': 1}, info[0][3])
